import gui.RoboGameFrame;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Random;

import algorithms.ManualControl;
import backend.GameLogic;
import backend.RobotAlgorithmIF;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * This class represents the execution code for the game. Running the main
 * method will display a window with a graphical representation of the game as
 * it runs.
 * 
 * The command-line options are of the form:
 * 
 * java ExecuteGame [-algorithms path1 [path2 path3...]] [-level intvalue] 
 * [-visible [true|false]] [-delay intvalue] [-exceptions [true|false]]
 * [-verbose [true|false]] [-package path] [-seed intvalue]
 * [-output [true|false]]
 * 
 * See the description at main for more information on the command line
 * arguments.
 * 
 * Created by Seth Sweep in fall, 2003. Heavily refactored by Matthew Justin,
 * Matthew Carlson, and Jon Quarfoth in fall, 2005.  Additional refactorings,
 * as well as bug fixing and addition of content added by Matthew Justin in
 * spring, 2005.
 * 
 * @author Seth Sweep
 * @author Matthew Justin
 * @author Matthew Carlson
 * @author Jon Quarfoth
 * @version 1.3
 * @see #main(String[])
 */
public class ExecuteGame {

    private static final String DEFAULT_PACKAGE = "<default>";

    private static final String ALGORITHM_PACKAGE_OPTION = "package";

    /**
     * List of all the legal arguments for use in main.
     */
    public static final List<String> legalArguments = legalArguments();

    /**
     * The execution code for the game. Displays a window with a graphical
     * representation of the game as it runs.
     * 
     * A variety of command line arguments are supported. The format for
     * arguments is:
     * 
     * -argument1 [value [value [...]]] -argument2 [value [value [...]]]
     * 
     * For arguments with boolean values, excluding the value in the option
     * produces the same result as -argument true
     * 
     * The legal arguments are:
     * 
     * algorithms: A list of the algorithms to include. Must include the package
     * path to the algorithms relative to the base package folder. Default is
     * one ManualControl.
     * 
     * package: The base package the algorithms are stored in. To use the
     * default (top level) package, use the value "&lt;default&gt;" (without the
     * quotes). The default value is "algorithms".
     * 
     * level: The level of the game to play. Currently, 1, 2, and 3 are the
     * supported levels. Any other value defaults to 1. Level 1 is a
     * non-randomized 51 x 51 grid consisting of buildings worth 1 point. Leve 2
     * is the same 51 x 51 grid, but randomizes the building values between 1
     * and 5, and adds additional random buildings. Level 3 uses randomized
     * building values and building locations, like level 2, but is a 61 x 81
     * grid. See the documentation for backend.PlayingField for more
     * information. The default value is 1.
     * 
     * visible: If set to true, displays the progress of the game graphically in
     * a window. Default value is true.
     * 
     * delay: Indicates the minimum amount of time, in miliseconds, each turn of
     * the game should take, where a turn consists of one move by each robot.
     * Values of 0 or less are treated as 0. This option can be used to slow
     * down the game to more clearly see the robots' progress. The default value
     * is 0.
     * 
     * exceptions: Set this to true to have exeptions thrown by the different
     * algorithms to be allowed to be thrown, causing the game to end. Having
     * this set to true is useful in debugging algorithms. Having this set to
     * false is useful in a tournament setting where you want the game to
     * complete, so the other robot(s) can continue to score points regardless
     * of whether other robots throw exceptions. Default value is false.
     * 
     * verbose: Set to true to display verbose game logging, by printing various
     * game data to standard out. Default value is false.
     * 
     * seed: Indicates the random seed to use for the random number generator in
     * the game. Two games run with the same seed are guaranteed to have the
     * same playing field configuration (building locations, building values,
     * and starting robot positions).  May be any <code>long</code> value.
     * Default is to use the default random constructor without a seed.
     * 
     * output: Whether to display the text output of the match.  Default is true.
     * 
     * Example command-line argument:
     * 
     * java ExecuteGame -visible true -verbose false -exceptions -delay 30
     * -algorithms ManualControl F03.NinjaPirate2 -package algorithms
     * 
     * If the arguments are malformed, or if any of the algorithms passed in
     * don't exist, don't have default constrors, thow an exception during
     * construction, or aren't valid algorithms, a message will be printed
     * stating the problem, after which the method will return without running
     * the game.
     * 
     * If an argument occurs more than once in the list of arguments, only the
     * last one applies. In the following example, the package will be set to
     * "algorithms", not the default package:
     * 
     * java ExecuteGame -package &lt;default&gt; -package algorithms
     * 
     * @param arguments The arguments to use. Each element of the array should
     *        be either an argument ("-verbose") or a value for the preceding
     *        argument ("false").
     */
    public static void main(String... arguments) {
        GameOptions options = defaultOptions();
        
        // Change this if you wish to use a different base package (and don't
        // want to specify this in the arguments).
        final String algorithmPackage = "algorithms";

        try {
            options = parseInput(arguments, algorithmPackage);
        } catch (IllegalArgumentException e) {
            System.out.println(e.getMessage());
            return;
        }
        
        /*
         * To add an algorithm to the game, place a new instance of it as part
         * of the array below. For example: RobotAlgorithm[] algorithms = { new
         * TeamSCoreAlgorithm(), new TabCompletionAlgorithm(), new
         * CowboyHackersAlgorithm(), };
         * 
         * Also make sure to uncomment the appropriate option below the array:
         * options.algorithms = algorithms
         */
        //RobotAlgorithmIF[] algorithms = {               
        // ****FALL 2003 ROBOTS****//
        //new Patsy(),
        // new StupidRandomAlgorithm(),
        // new TronClear(),
        // new ScadBot7(),
        // new RolandWrinkle(),
        // new FourDimensionalShadow(),
        // new SupremeRobot(),
        // new Win4(),
        // new NinjaPirate2(),
        // new ProjectKEA2Columns(),
        // new FriendlyBot(),
        // new McMuffinBot(),
        // new EugBot(),
        // new QuasimodoBot4(),
        // new QuasimodoBot2Point2(),
        // new CaptainUnderPants(),
        // ****END FALL 2003 ROBOTS****//
        //
        // ****FALL 2004 ROBOTS****//
        // new StupidRandomAlgorithm(),
        // new StellarCartographer(),
        // new runAway(),
        // ****END FALL 2004 ROBOTS****//
        //        
        // ****TEST ROBOTS (RUN ALL 4 FOR A FUN PATTERN)****
//         new PredictableTestRobot(),
//         new PredictableTestRobot(),
//         new PredictableTestRobot(),
//         new PredictableTestRobot(),
        // ****END TEST ROBOTS****
        //        new SorcererBot(),
//        new GrandPooBot()
//        		new ChameleonBot(),
//        		new PredictableTestRobot()
//        };

        /*
         * To change game options (overriding the command-line arguments),
         * uncomment out the appropriate fields and give them the desired
         * values.
         */
//        options.algorithms = algorithms;
//        options.verbose = false;
//        options.visible = true;
//        options.exceptions = false;
//        options.level = 1;
//        options.delay = 0;

        
        // added by Daniel Selifonov for creation of a double-round-robin tournament
        RobotAlgorithmIF[] algorithms = options.algorithms;
        //int tournament_rounds = algorithms.length * (algorithms.length - 1);
        int current_round = 1;
        options.output = false;
        int[] score_buffer = new int[algorithms.length];
        int[] wins_buffer = new int[algorithms.length];
        for (int i = 0; i < score_buffer.length; i++)
        {
        	score_buffer[i] = 0;
        	wins_buffer[i] = 0;
        }
        for (int i = 0; i < algorithms.length; i++)
        {
        	for (int j = 0; j < algorithms.length; j++)
        	{
        		if (i != j)
        		{
        			RobotAlgorithmIF[] in_round = new RobotAlgorithmIF[2];
        			in_round[0] = algorithms[i];
        			in_round[1] = algorithms[j];
        			options.algorithms = in_round;
        			
        			System.out.println("Tournament Round " + current_round + ": " + in_round[0].getName() + " vs. " + in_round[1].getName());
        			
        			int[] round_scores = executeGame(options);
        			score_buffer[i] += round_scores[0];
        			score_buffer[j] += round_scores[1];
        			
        			String result = "Robot ";
        			if (round_scores[0] > round_scores[1])
        			{
        				wins_buffer[i]++;
        				result = in_round[0].getName() + " won. ";
        			}
        			else if (round_scores[0] < round_scores[1])
        			{
        				wins_buffer[j]++;
        				result = in_round[1].getName() + " won. ";
        			}
        			else
        			{
        				wins_buffer[i]++;
        				wins_buffer[j]++;
        				result = "Round tie! ";
        			}
        			result = result + round_scores[0] + " to " + round_scores[1] + " points captured.\n";
        			System.out.println(result);
            		algorithms = reconstructRobots(algorithms);
        			current_round++;
        			try
					{
						Thread.sleep(1000);
					}
					catch (InterruptedException e)
					{
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
        		}
        	}
        }
        
        // print double-round-robin tournament results
        System.out.println("Final Results:");
        System.out.println("\nRobot\t\t|\tRounds Won\t|\tTotal Score");
        for (int i = 0; i < score_buffer.length; i++)
        {
        	System.out.println(algorithms[i].getName() + "\t|\t" + wins_buffer[i] + "\t|\t" + score_buffer[i]);
        }
        
        //executeGame(options);
    }
    
    private static RobotAlgorithmIF[] reconstructRobots(RobotAlgorithmIF[] input)
    {
    	RobotAlgorithmIF[] output = new RobotAlgorithmIF[input.length];
    	for (int i = 0; i < input.length; i++)
    	{
    		try
			{
				output[i] = input[i].getClass().newInstance();
			}
			catch (InstantiationException e)
			{
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			catch (IllegalAccessException e)
			{
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
    	}
    	
    	return output;
    }
    
    private static GameOptions defaultOptions() {
        RobotAlgorithmIF[] algorithms = { new ManualControl() };

        final boolean verbose = false;
        final boolean visible = true;
        final boolean throwBotExceptions = false;
        final int level = 1;
        final int delay = 0;
        final Long seed = null;
        final boolean output = true;

        GameOptions defaults = new GameOptions(algorithms, level, visible,
                delay, throwBotExceptions, verbose, seed, output);
        return defaults;
    }

    private static GameOptions parseInput(String[] args, String algorithmPackage) {
        if (args.length == 0) {
            return defaultOptions();
        }
        Map<String, List<String>> argumentMap = CommandLineParser.parse(args);

        algorithmPackage = getAlgorithmPackage(algorithmPackage, argumentMap);
        argumentMap.remove(ALGORITHM_PACKAGE_OPTION);

        GameOptions result = toGameOptions(algorithmPackage, argumentMap);

        setUnspecifiedValuesToDefaults(result);
        return result;
    }

    private static String getAlgorithmPackage(String algorithmPackage,
            Map<String, List<String>> argumentMap) {
        List<String> packageList = argumentMap.get(ALGORITHM_PACKAGE_OPTION);
        if (packageList == null) {
            return algorithmPackage;
        }
        if (packageList.size() != 1) {
            throw new IllegalArgumentException(
                    "Illegal arguments for algorithm package");
        }
        return packageList.get(0);
    }

    private static void setUnspecifiedValuesToDefaults(GameOptions result) {
        GameOptions defaults = defaultOptions();
        for (String argument : GameOptions.LEGAL_ARGUMENTS) {
            if (result.getOption(argument) == null) {
                result.setOptionToValue(argument, defaults.getOption(argument));
            }
        }
    }

    private static GameOptions toGameOptions(String algorithmPackage,
            Map<String, List<String>> argumentMap) {
        List<String> algorithms = argumentMap.get("algorithms");
        if (algorithms != null) {
            for (int i = 0; i < algorithms.size(); i++) {
                String algorithm = algorithms.get(i);
                if (!algorithmPackage.equals(DEFAULT_PACKAGE)) {
                    algorithm = algorithmPackage + "." + algorithm;
                    algorithms.set(i, algorithm);
                }
            }
        }
        
        GameOptions result = new GameOptions();
        for (Map.Entry<String, List<String>> entry : argumentMap.entrySet()) {
            result.setOption(entry.getKey(), entry.getValue());
        }
        return result;
    }
    
    private static List<String> legalArguments() {
        List<String> result = new ArrayList<String>(GameOptions.LEGAL_ARGUMENTS);
        result.add(ALGORITHM_PACKAGE_OPTION);
        return Collections.unmodifiableList(result);
    }

    /**
     * Executes the game, with the given set of game options. Any options set to
     * null will use their default settings.
     * 
     * @param options The game options to use.
     */
    public static int[] executeGame(GameOptions options) {
        return executeGame(options.algorithms, options.level, options.visible,
                options.delay, options.exceptions, options.verbose,
                options.seed, options.output);
    }

    /**
     * The execution code for the game.
     * 
     * @param algorithms The algorithms to use.
     * @param level The level of the game to use. Currently 1, 2, and 3 are
     *        supported. See the documentation of <code>PlayingField</code>
     *        for more information.
     * @param visible Whether the progress of the game should be displayed
     *        graphically in a window.
     * @param delay The minimum amount of time (in miliseconds) each turn should
     *        take.
     * @param throwBotExceptions Whether to throw bot exceptions, or quietly
     *        catch them.
     * @param verbose Whether verbose logging should be used.
     * @param seed The random seed to use (or null to use the default random).
     * @param output Whether to display the output of the game.
     */
    public static int[] executeGame(RobotAlgorithmIF[] algorithms,
            final int level, final boolean visible, int delay,
            boolean throwBotExceptions, final boolean verbose, Long seed,
            boolean output) {
        Random random = (seed == null ? new Random() : new Random(seed));
        GameLogic gl = new GameLogic(algorithms, level, random);
        if (visible) {
            RoboGameFrame kungFooRobots = new RoboGameFrame(gl);
            kungFooRobots.setVisible(true);
        }
        gl.setVerbose(verbose);
        if (output) {
            printInitialGameInformation(GameLogic.VERSION, level, algorithms.length);
        }
        gl.setTurnDelay(delay);
        gl.setThrowBotExceptions(throwBotExceptions);
        int[] result = gl.execute();
        if (output) {
            printEndGameReport(gl);
        }
        return result;
    }

    private static void printInitialGameInformation(String version, int level,
            int numAlgorithms) {
        System.out.println("Kung Foo Robots -- Ver. " + version);
        System.out.println("----------------------------------------");
        System.out.println("Initializing the playing field at level " + level
                + "...");
        System.out.println("Starting game with " + numAlgorithms
                + " robot(s)...");
    }

    /**
     * Outputs each robot's scores and declares the winner.
     */
    private static void printEndGameReport(GameLogic logic) {
        System.out.println(logic.endGameReport());
    }
}
