import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Formatter;
import java.util.List;

import backend.RobotAlgorithmIF;

/**
 * Class comprising the various possible options for a game.
 * 
 * Null values for any of the options indicates that the default values should
 * be used.
 */
public final class GameOptions {

    /**
     * List of the legal arguments for the getters and setters.
     */
    public static final List<String> LEGAL_ARGUMENTS = legalArguments();

    /**
     * List of algorithms to use.
     */
    public RobotAlgorithmIF[] algorithms;

    /**
     * The game level.
     */
    public Integer level;

    /**
     * Whether to display the game progress graphically.
     */
    public Boolean visible;

    /**
     * The minimum amount of delay per game turn (in miliseconds).
     */
    public Integer delay;

    /**
     * Whether to throw exceptions thrown by algorithms.  If false, indicates
     * that the exceptions should be caught.
     */
    public Boolean exceptions;

    /**
     * Whether to use verbose logging.
     */
    public Boolean verbose;

    /**
     * The random seed to use for the game.
     */
    public Long seed;

    /**
     * Whether to display the output of the match.
     */
    public Boolean output;

    /**
     * Creates a new instance of this class, with all the options set to null.
     */
    public GameOptions() {
    }

    /**
     * Creates a new instance of this class, with the given options.  Sets the
     * approprate fields to the given values.
     * 
     * @param algorithms The algorithms to use.
     * @param level The game level.
     * @param visible Whether to display the game graphically.
     * @param delay The minimum amount of delay each turn.
     * @param exceptions Whether to propogate exceptions.
     * @param verbose Whether to use verbose logging.
     */
    public GameOptions(RobotAlgorithmIF[] algorithms, Integer level,
            Boolean visible, Integer delay, Boolean exceptions,
            Boolean verbose, Long seed, Boolean output) {
        this.algorithms = algorithms;
        this.level = level;
        this.visible = visible;
        this.delay = delay;
        this.exceptions = exceptions;
        this.verbose = verbose;
        this.seed = seed;
        this.output = output;
    }

    private static List<String> legalArguments() {
        List<String> result = new ArrayList<String>();
        for (Field field : GameOptions.class.getDeclaredFields()) {
            if (validField(field)) {
                result.add(field.getName());
            }
        }
        return Collections.unmodifiableList(result);
    }

    private static boolean validField(Field field) {
        int modifiers = field.getModifiers();
        boolean result = true;
        result = result && !Modifier.isFinal(modifiers);
        result = result && !Modifier.isStatic(modifiers);
        result = result && Modifier.isPublic(modifiers);
        return result;
    }

    private IllegalArgumentException toIllegalArgumentException(Exception e) {
        return new IllegalArgumentException("A " + e.getClass().getSimpleName()
                + " was thrown", e);
    }

    /**
     * Sets the option with the given name to the given value.
     * 
     * @param option The name of the option to set.
     * @param value The value to set the option to.
     * @throws IllegalArgumentException Thrown if <code>option</code> is not a
     *         valid option, or if the value isn't valid for the given option.
     */
    public void setOptionToValue(String option, Object value) {
        if (!(LEGAL_ARGUMENTS.contains(option))) {
            throw new IllegalArgumentException(option
                    + " is not a valid argument");
        }
        Field field = getField(option);
        try {
            field.set(this, value);
        } catch (IllegalAccessException e) {
            toIllegalArgumentException(e);
        } catch (IllegalArgumentException e) {
            throw new IllegalArgumentException(value
                    + " is an illegal argument for " + option);
        }
    }

    /**
     * Sets the option to the given values. The values will be translated from
     * their string representation to their appropriate value. This is the same
     * as setOptions(option, String... values); see that method for more
     * information.
     * 
     * @param option The option to set.
     * @param values The values to set the option to.
     * @throws IllegalArgumentException Thrown if <option>option</option> isn't
     *         a valid option, or if the value list can't be converted to the
     *         appropriate option type (such as when values.size() != 1 for an
     *         integer option).
     */
    public void setOption(String option, List<String> values) {
        setOption(option, values.toArray(new String[values.size()]));
    }

    /**
     * Sets the option to the given values. The values will be translated from
     * their string representation to their appropriate value.
     * 
     * Legal String representations of values:
     * 
     * For boolean values, the legal values are arrays containing no items
     * (intrepreted as <code>true</code>), as well as one-element arrays
     * containing either "true" or "false".
     * 
     * Legal integer values are one-element arrays containing any String that is
     * a legal argument for <code>Integer.parseInt(String)</code>
     * 
     * Legal RobotAlgorithmIF[] values are arrays of length 1 or more, each
     * element of which must be the class name (including package path)
     * corresponding to a valid non-abstract RobotAlgorithmIF[] object with a
     * no-element constructor that doesn't throw an exception when invoked. An
     * example would be "algorithms.ManualControl".
     * 
     * @param option The option to set.
     * @param values The values to set the option to.
     * @throws IllegalArgumentException Thrown if <option>option</option> isn't
     *         a valid option, or if the value list can't be converted to the
     *         appropriate option type (such as when values.length != 1 for an
     *         integer option).
     */
    public void setOption(String option, String... values) {
        if (!(LEGAL_ARGUMENTS.contains(option))) {
            throw new IllegalArgumentException(option
                    + " is not a valid argument");
        }
        Field field = getField(option);
        Object value = null;
        if (field.getType().equals(Boolean.class)) {
            value = toBoolean(option, values);
        } else if (field.getType().equals(Integer.class)) {
            value = toInt(option, values);
        } else if (field.getType().equals(Long.class)) {
            value = toLong(option, values);
        } else if (field.getType().equals(RobotAlgorithmIF[].class)) {
            value = toAlgorithmArray(option, values);
        } else {
            throw new UnsupportedOperationException("The class type "
                    + field.getType() + " should not occur in GameOptions");
        }
        setOptionToValue(option, value);
    }

    /**
     * Returns the value of the named option.
     * 
     * @param option The name of the option to return the value of.
     * @return The value of the option.
     * @throws IllegalArgumentException Thrown if <option>option</option> isn't
     *         a valid option.
     */
    public Object getOption(String option) {
        if (!(LEGAL_ARGUMENTS.contains(option))) {
            throw new IllegalArgumentException(option
                    + " is not a valid argument");
        }
        return fieldValue(getField(option));
    }

    private Field getField(String option) {
        Field field;
        try {
            field = getClass().getDeclaredField(option);
        } catch (SecurityException e) {
            throw toIllegalArgumentException(e);
        } catch (NoSuchFieldException e) {
            throw new IllegalArgumentException("An illegal option was passed");
        }
        return field;
    }

    private Object fieldValue(Field field) {
        try {
            return field.get(this);
        } catch (IllegalAccessException e) {
            throw toIllegalArgumentException(e);
        }
    }

    private boolean toBoolean(String option, String[] values) {
        boolean value;
        if (values.length > 1) {
            throw new IllegalArgumentException(option
                    + " can have at most one parameter");
        } else if (values.length == 0) {
            value = true;
        } else {
            String valueString = values[0];
            if (!(valueString.equalsIgnoreCase(""+false) || valueString
                    .equalsIgnoreCase("" + true))) {
                throw new IllegalArgumentException(option
                        + " can only be true or false");
            }
            value = Boolean.valueOf(valueString);
        }
        return value;
    }

    private int toInt(String option, String[] values) {
        if (values.length != 1) {
            throw new IllegalArgumentException(option
                    + " must have exactly one parameter");
        }
        int result;
        try {
            result = Integer.parseInt(values[0]);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(values[0]
                    + " is an illegal argument for " + option);
        }
        return result;
    }

    private long toLong(String option, String[] values) {
        if (values.length != 1) {
            throw new IllegalArgumentException(option
                    + " must have exactly one parameter");
        }
        long result;
        try {
            result = Long.parseLong(values[0]);
        } catch (NumberFormatException e) {
            throw new IllegalArgumentException(values[0]
                    + " is an illegal argument for " + option);
        }
        return result;
    }

    private RobotAlgorithmIF[] toAlgorithmArray(String option, String[] values) {
        if (values.length == 0) {
            throw new IllegalArgumentException(option
                    + " must have at least one parameter");
        }
        RobotAlgorithmIF[] result = new RobotAlgorithmIF[values.length];
        for (int i = 0; i < values.length; i++) {
            result[i] = getAlgorithm(values[i]);
        }
        return result;
    }

    private RobotAlgorithmIF getAlgorithm(String classPath) {
        Class<RobotAlgorithmIF> algClass = getClass(classPath);
        Constructor<RobotAlgorithmIF> constructor = getConstructor(algClass,
                classPath);
        RobotAlgorithmIF algorithm = constructObject(constructor, classPath);
        return algorithm;
    }

    private <T> T constructObject(Constructor<T> constructor, String classPath) {
        T result;
        try {
            result = constructor.newInstance((Object[]) null);
        } catch (IllegalArgumentException e) {
            // wrong number of arguments; can't happen.
            throw toIllegalArgumentException(e);
        } catch (InstantiationException e) {
            throw new IllegalArgumentException("The algorithm named "
                    + classPath + " was an abstract class");
        } catch (IllegalAccessException e) {
            // no access to constructor; can't happen (I think).
            throw toIllegalArgumentException(e);
        } catch (InvocationTargetException e) {
            throw new RuntimeException("The algorithm named " + classPath
                    + " threw an exception: " + e.getClass().getSimpleName(), e
                    .getCause());
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    private Class<RobotAlgorithmIF> getClass(String classPath) {
        Class<?> result;
        try {
            result = Class.forName(classPath);
        } catch (ClassNotFoundException e) {
            throw new IllegalArgumentException("No algorithm named "
                    + classPath + " was found");
        }
        if (!(RobotAlgorithmIF.class.isAssignableFrom(result))) {
            throw new IllegalArgumentException("The object named " + classPath
                    + " does not implement "
                    + RobotAlgorithmIF.class.getSimpleName());
        }
        return (Class<RobotAlgorithmIF>)result;
    }

    private <T> Constructor<T> getConstructor(Class<T> clazz, String classPath) {
        try {
            return clazz.getConstructor(new Class[0]);
        } catch (SecurityException e) {
            throw toIllegalArgumentException(e);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException("The algorithm named "
                    + classPath + " did not have a no-argument constructor");
        }
    }

    /**
     * Returns a string representation of this class.  Displays the various
     * options, and their values.
     * 
     * @return A string representation of this class.
     */
    public String toString() {
        StringBuilder result = new StringBuilder();
        result.append("GameOptions:");
        Formatter formatter = new Formatter(result);
        for (String string : LEGAL_ARGUMENTS) {
            Field field = getField(string);
            Object fieldValue = fieldValue(field);
            if (fieldValue != null && fieldValue.getClass().isArray()) {
                fieldValue = Arrays.toString((Object[])fieldValue);
            }
            formatter.format(" %s: %s;", field.getName(), fieldValue);
        }
        result.deleteCharAt(result.length() - 1);
        return result.toString();
    }
}