package algorithms;

import java.util.*;

import backend.Move;
import backend.RobotAlgorithmIF;
import backend.SensorGroup;
import backend.Sensor.BuildingOwner;
import backend.Sensor.CellType;

/**
 * @author Nathan Dahlberg, Fernando, Jeff
 * on $Date: 2007/09/26 15:37:24 $
 * @version $Revision: 1.1 $
 *
 */
public class ChameleonBot implements RobotAlgorithmIF{

    /**
     * Keeps track of distances moved in each direction while capturing
     */
    private HashMap<Integer,Integer> distances = new HashMap<Integer,Integer>();
    private int direction = 0;
    
    /**
     * Queue of predetermined moves, added to capture certain squares
     */
    private LinkedList<Move> setMoves = new LinkedList<Move>();
    
    /**
     * The width and length the squares our bot makes
     */
    private int optimal_width = 24;
    private int optimal_length = 50;
    
    /**
     * Number of moves spent NOT capturing something in a row
     */
    private int useless_moves = 0;
    
    /**
     * Keeps track of ownership of passing buildings, so we can capture key spots
     */
    private BuildingOwner right = BuildingOwner.ADVERSARY;
    private BuildingOwner left = BuildingOwner.ADVERSARY;
    private BuildingOwner aheadright = BuildingOwner.ADVERSARY;
    private BuildingOwner aheadleft = BuildingOwner.ADVERSARY;
    
    /**
     * Constructor, adds Keys to the hashmap
     */
    public ChameleonBot() {
        distances.put(0, 0);
        distances.put(1, 0);
        distances.put(2, 0);
        distances.put(3, 0);
    }
    
    public Move getMove(SensorGroup sensors) {
        
        //if a breadcrumb is ahead, complete capture and restart algorithm
        if (sensors.aheadOne().hasMyCrumbs()) {
            distances.put(0, 0);
            distances.put(1, 0);
            distances.put(2, 0);
            distances.put(3, 0);
            setMoves.clear();
            return Move.AHEAD_WITH_CRUMBS;
        }
        
        //if we are capture a set piece to mess up our opponent...
        if (!setMoves.isEmpty()) {
            return setMoves.remove();
        }
        //if there is a breadcrumb on this space, we must interfere with our opponent!
        if (sensors.center().hasAdversarysCrumbs()&&((sensors.left().getCellType()==CellType.STREET)||
                (sensors.right().getCellType()==CellType.STREET))) {
            optimal_width=20;
            optimal_length=30;
            if (sensors.aheadRight().getCellType()==CellType.BUILDING) {
                aheadright = sensors.aheadRight().getOwner();
            } else {
                aheadright = BuildingOwner.ADVERSARY;
            }
            if (sensors.right().getCellType()==CellType.BUILDING) {
                aheadleft = sensors.aheadLeft().getOwner();
            } else {
                aheadleft = BuildingOwner.ADVERSARY;
            }
            
            //depending on free spaces, setup capture technique
            if (aheadleft==BuildingOwner.UNCLAIMED&&left==BuildingOwner.UNCLAIMED) {
                //capture left 2 squares
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(4,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                resetSpaces();
                return getMove(sensors);
            }
            
            if (aheadright==BuildingOwner.UNCLAIMED&&right==BuildingOwner.UNCLAIMED) {
                //capture right 2 squares
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(4,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                resetSpaces();
                return getMove(sensors);
            }
            
            
            if (aheadright==BuildingOwner.UNCLAIMED&&aheadleft==BuildingOwner.UNCLAIMED) {
                //capture ahead 2 squares
                queueAdd(0,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(4,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                resetSpaces();
                return getMove(sensors);
            }
            if (aheadright==BuildingOwner.UNCLAIMED) {
                //capture ahead right square
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                resetSpaces();
                return getMove(sensors);
            }
            if (aheadleft==BuildingOwner.UNCLAIMED) {
                //capture ahead left square
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                resetSpaces();
                return getMove(sensors);
            }
            if (right==BuildingOwner.UNCLAIMED) {
                //capture behind right square
                queueAdd(0,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                queueAdd(2,Move.TURN_RIGHT);
                resetSpaces();
                return getMove(sensors);
            }
            if (left==BuildingOwner.UNCLAIMED) {
                //capture behind left square
                queueAdd(0,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                queueAdd(2,Move.TURN_LEFT);
                resetSpaces();
                return getMove(sensors);
            }
        }
        //if there is a breadcrumb ahead, we must prepare to interfere with our opponent!
        if (sensors.aheadOne().hasAdversarysCrumbs()&&((sensors.left().getCellType()==CellType.BUILDING)||
                                                      (sensors.right().getCellType()==CellType.BUILDING))) {
            if (sensors.right().getCellType()==CellType.BUILDING) {
                right = sensors.right().getOwner();
            } else {
                right = BuildingOwner.ADVERSARY;
            }
            if (sensors.left().getCellType()==CellType.BUILDING) {
                left = sensors.left().getOwner();
            } else {
                left = BuildingOwner.ADVERSARY;
            }
        }
        
        //if we are capturing stuff...
        if(distances.get(0)!=0) {            
            return captureBuildings(sensors);
        }        
        //if we are searching for stuff to capture...        
        return moveSearch(sensors);
    }

    /**
     * This method is called when we are capturing an area, it returns the next valid move
     * within the area of it's optimal_width and optimal_length.
     * 
     * @param sensors
     * @return move to make
     */
    private Move captureBuildings(SensorGroup sensors) {
        if (sensors.aheadRight().getOwner()!=BuildingOwner.UNCLAIMED
            ||sensors.aheadOne().getCellType()==CellType.WALL) {
            return turnRight();
        }
         
        if (direction==0||direction==2) {
            if (direction>0&&distances.get(direction-1)<optimal_length&&
                sensors.aheadLeft().getCellType()==CellType.BUILDING&&
                sensors.aheadLeft().getOwner()==BuildingOwner.UNCLAIMED) {
                return turnLeft();
            }
            if (distances.get(direction)<optimal_width) {
                distances.put(direction, distances.get(direction)+1);
                return Move.AHEAD_WITH_CRUMBS;
            }
            return turnRight();
        }            
        if (distances.get(direction)<optimal_length) {
            if (distances.get(direction-1)<optimal_width&&
                sensors.aheadLeft().getCellType()==CellType.BUILDING&&
                sensors.aheadLeft().getOwner()==BuildingOwner.UNCLAIMED) {
                return turnLeft();
            }
           distances.put(direction, distances.get(direction)+1);
            return Move.AHEAD_WITH_CRUMBS;
        }
        return turnRight();
    }

    /**
     * This method is called when no predetermined moves are running and there are no
     * obvious places to capture.  It circles the board, then cuts across it until it finds
     * a valid move.
     * 
     * @param sensors
     * @return next move
     */
    private Move moveSearch(SensorGroup sensors) {
        //if we find something, start capture
        if ((sensors.aheadRight().getCellType()==CellType.BUILDING
            &&sensors.aheadRight().getOwner()==BuildingOwner.UNCLAIMED)
             ) {
            direction=0;
            distances.put(0, 1);
            useless_moves = 0;
            return Move.AHEAD_WITH_CRUMBS;
        }        
        //if we don't find something, this move was useless
        useless_moves++;        
        //if we have gone through the middle twice, divide into quadrants and search
        if (useless_moves>1000) {
            if (sensors.farAhead().getCellType()==CellType.WALL&&sensors.farSensorDistance()==12) {
                return Move.TURN_RIGHT;
            }
        }       
        //if we have circled and found nothing, go through the middle both directions
        if (useless_moves>205) {
            if (sensors.farAhead().getCellType()==CellType.WALL&&sensors.farSensorDistance()==26) {
                return Move.TURN_RIGHT;
            }
        }        
        //if we are not yet bored, circle the whole thing
        if (sensors.aheadOne().getCellType()==CellType.WALL) {
            return Move.TURN_RIGHT;
        }            
        return Move.AHEAD_WITH_CRUMBS;
    }

    /**
     * Method to make right turns that keeps the directions correct
     * 
     * @return a right turn
     */
    private Move turnRight() {
        direction++;
        if (direction>3) {
            direction=0;
        }
        return Move.TURN_RIGHT;
    }
    
    /**
     * Method to make left turns that keeps the directions correct
     * 
     * @return a left turn
     */
    private Move turnLeft() {
        direction--;
        if (direction<0) {
            direction=3;
        }
        return Move.TURN_LEFT;
    }
    
    /**
     * After setting up the capture of a key space, reset the known squares
     */
    private void resetSpaces() {
        right = BuildingOwner.ADVERSARY;
        left = BuildingOwner.ADVERSARY;
        aheadright = BuildingOwner.ADVERSARY;
        aheadleft = BuildingOwner.ADVERSARY;
    }
    
    /**
     * Add forward number of Move ahead to predetermined moves, then turn in direction
     * 
     * @param forward
     * @param moveDirection
     */
    private void queueAdd(int forward, Move moveDirection) {
        for (int i=0; forward>i; i++) {
            setMoves.add(Move.AHEAD_WITH_CRUMBS);
        }
        setMoves.add(moveDirection);
    }
    
    public String getName() {
        return "ChameleonBot";
    }

}
