package algorithms;

import backend.Move;
import backend.RobotAlgorithmIF;
import backend.SensorGroup;

/**
 * This robot follows narrow rectangular paths as a capture strategy.  It starts
 * by following six tall, narrow rectangular paths, leaving small gaps between these
 * regions to be captured later, then switches the orientation of its capture by making 
 * short, wide rectangles.  Eventually, the rectangles will have enclosed every building
 * in the entire game.
 *  
 * The idea behind this algorithm is to prevent the opposition from trying
 * to capture large numbers of buildings at one time by capturing small regions 
 * quickly, foiling enemy capture stategies.
 * 
 * @author Group 1
 */
public class MightyBot implements RobotAlgorithmIF {
    
       /**
        * The width of the rectangular path taken by the bot.
        */
        private final static int WIDTH = 6;
        
        /**
         * The height of the rectangular path taken by the bot.
         */
        private final static int LENGTH = 46;
        
        /**
         * The space between each additional path and the previous.
         */
        private final static int SKIP = 3;
        
        /**
         * The number of moves it takes for the bot to make a right turn.
         * In the case of all bot algorithms in this project, this number 
         * will always be equal to 1.
         */
        private final static int TURN_COST = 1;
        
        /**
         * Each 'turn' is a count of the number of moves this bot has taken
         * up to that turn.  To make a full rectangular path, there are four 
         * corners where the robot must turn, named below.  The numbering of 
         * the turns is based on when the robot makes the turn.  So turn 1 is
         * the first turn the robot makes when making the rectangle.
         */
        private int turn1 = WIDTH, 
        			turn2 = turn1 + LENGTH + TURN_COST, 
        			turn3  = turn2 + WIDTH + TURN_COST, 
        			turn4 = turn3 + LENGTH + TURN_COST; 
        
        /**
         * The numbers of rectangular paths taken by the bot of a certain
         * orientation.  The bot alternates between vertical orientation and
         * horizontal orientation.
         */
        private int iterations = 0;
        
        /**
         * A count of the number of game turns.
         */
        private int gameTurns = 0;

    /**
     * Moves the robot according to the algorithm described above.  If the robot
     * would collide with another robot, it will instead evade by turning right and will 
     * resume capturing the current rectangle.
     */
    public Move getMove(SensorGroup sensors) {
        if (collisionImminent(sensors) && sensors.right().getCellType() == sensors.aheadOne().getCellType()) {
            if (gameTurns < turn1)
            	gameTurns = turn1;
            else if (gameTurns < turn2)
            	gameTurns = turn2;
            else if (gameTurns < turn3)
            	gameTurns = turn3;
            else
            	gameTurns = turn4;
        }
        
        
        if (iterations < 6)
    	{
    		if ( gameTurns == turn1 ||gameTurns == turn2 ||gameTurns == turn3 || gameTurns == turn4)
    		{
    			gameTurns++;
    			return Move.TURN_RIGHT;
    		}
    		else if (gameTurns < turn4 + WIDTH + SKIP + 1) 
    		{
    			gameTurns++;
    			return Move.AHEAD_WITH_CRUMBS;
    		}
    		else
    		{
    			
    			gameTurns = 0;
    			iterations++;
    			return Move.AHEAD_WITH_CRUMBS;
    		}
    	}
        else {
            iterations = 0;
            return Move.TURN_RIGHT;
        }
    }
    
    /**
     * Checks to see if this bot is in danger of colliding with another bot.  A bot is considered a threat
     * if it is within two spaces directly in front of this bot.
     * @param sensors 
     * @return a boolean value
     */
    private boolean collisionImminent(SensorGroup sensors) {
    	if (sensors.farSensorDistance() == 1 && sensors.farAhead().hasAdversary())
    		return true;
    	else 
    		return false;
    }

    /**
     * Returns the name of this robot.
     * @return <code>this robot's name</code>
     */
    public String getName() {
        return "MightyBot";
    }
}
