package backend;

/**
 * Algorithms and Compubility -- Project 2
 * 
 * This class defines a claimable cell (building). Each building has an owner
 * (or is unowned), a point value, and a flag indicating whether this building
 * was "checked" (used by the enclosure scanning algorithm for awarding points).
 * 
 * @author Seth Sweep
 */
public class BuildingCell implements Cell {
    /**
     * The default number of points (1).
     */
    public static final int DEFAULT_POINTS = 1;

    /**
     * How many points the cell is worth. Points may be negative! (Later)
     */
    private int _points;

    /**
     * Which robot owns the cell. (null = nobody)
     */
    private Robot _owner = null;

    /**
     * Has this been checked? Used for awarding points.
     */
    private boolean _checked = false;

    private BuildingCaptureListener _listener;

    /**
     * Initialize the building cell with the default point value;
     */
    public BuildingCell() {
        this(DEFAULT_POINTS);
    }

    /**
     * Initialize the building cell with a given number of points.
     * 
     * @param points The point value of the building cell.
     */
    public BuildingCell(int points) {
        setPoints(points);
    }

    /**
     * Sets the point value of this building.
     * 
     * @param points The value to set.
     */
    public void setPoints(int points) {
        _points = points;
    }

    /**
     * Returns the point value of this building.
     * 
     * @return the point value of this building.
     */
    public int getPoints() {
        return _points;
    }

    /**
     * Sets the owner of this Cell.  If owner is null, then the building is unowned.
     * 
     * @param owner The owner to set.
     */
    public void setOwner(Robot owner) {
        _owner = owner;
        if (_listener != null) {
            _listener.fireBuildingChanged(this);
        }
    }

    /**
     * Return the owner of this building, or null if the building is unowned.
     * 
     * @return the owner of this building.
     */
    public Robot getOwner() {
        return _owner;
    }

    /**
     * Returns true iff this building is owned
     * 
     * @return Whether this building is owned.
     */
    public boolean isOwned() {
        return _owner != null;
    }


    /**
     * Returns true iff this building is equal to another object.  Returns true
     * only if the other object is a BuildingCell with the same point value
     * and owner as this one.
     * 
     * @param object The object to compare this to.
     * @return Whether the obect is equal to this.
     */
    @Override
    public boolean equals(Object object) {
        if (!(object instanceof BuildingCell)) {
            return false;
        }
        BuildingCell b = (BuildingCell) object;
        return (_points == b._points) && areEqual(_owner, b._owner);
    }

    private boolean areEqual(Object o1, Object o2) {
        return (o1 == null ? o2 == null : o1.equals(o2));
    }

    /**
     * Returns the hash code of this object.
     * 
     * @return This building's hash code.
     */
    @Override
    public int hashCode() {
        int result = 17;
        result = 37 * result + _points;
        result = 37 * result + (_owner == null ? 0 : _owner.hashCode());
        return result;
    }

    /**
     * Returns a String representation of this building.
     * 
     * @return A String representation of this cell.
     */
    public String toString() {
        return "Building cell: points: " + _points + "; owner: " + _owner;
    }

    /**
     * Returns true iff this building is checked.
     * 
     * @return Whether this cell is checked.
     */
    public boolean isChecked() {
        return _checked;
    }

    /**
     * Sets the checked value of this field.
     * 
     * @param checked The value to set
     */
    public void setChecked(boolean checked) {
        _checked = checked;
    }

    /**
     * Returns true if this cell is an obstacle.  Always returns true.
     * 
     * @return Whether this cell is an obstacle
     */
    public boolean isObstacle() {
        return true;
    }

    /**
     * Returns the message to be logged when a robot attempts to pass through
     * this when it's an obstacle.
     * 
     * @return The obstacle message.
     */
    public String getObstacleMessage() {
        return "Game Logic: Robot bangs itself against a building.";
    }

    /**
     * Returns the type of this cell.  Returns CellType.BUILDING.
     * 
     * @return The type of this cell.
     */
    public Sensor.CellType cellType() {
        return Sensor.CellType.BUILDING;
    }

    /**
     * Sets the listener to use.
     * 
     * @param listener The listener to use.
     */
    public void setListener(BuildingCaptureListener listener) {
        _listener = listener;
    }
}
