package backend;

import java.awt.Point;
import java.util.Random;

/**
 * An enum representing one of the four cardinal directions (north, south, east,
 * and west).
 * 
 * @author Matthew Justin
 */
public enum Direction {
    NORTH {
        public Direction left() {
            return WEST;
        }

        public Direction right() {
            return EAST;
        }

        public Direction back() {
            return SOUTH;
        }

        public Point getOffsetPoint(Point point, int foreDistance, int rightDistance) {
            int x = point.x + rightDistance;
            int y = point.y - foreDistance;
            return new Point(x, y);
        }
    },
    EAST {
        public Direction left() {
            return NORTH;
        }

        public Direction right() {
            return SOUTH;
        }

        public Direction back() {
            return WEST;
        }

        public Point getOffsetPoint(Point point, int foreDistance, int rightDistance) {
            int x = point.x + foreDistance;
            int y = point.y + rightDistance;
            return new Point(x, y);
        }
    },
    SOUTH {
        public Direction left() {
            return EAST;
        }

        public Direction right() {
            return WEST;
        }

        public Direction back() {
            return NORTH;
        }

        public Point getOffsetPoint(Point point, int foreDistance, int rightDistance) {
            int x = point.x - rightDistance;
            int y = point.y + foreDistance;
            return new Point(x, y);
        }
    },
    WEST {
        public Direction left() {
            return SOUTH;
        }

        public Direction right() {
            return NORTH;
        }

        public Direction back() {
            return EAST;
        }

        public Point getOffsetPoint(Point point, int foreDistance, int rightDistance) {
            int x = point.x - foreDistance;
            int y = point.y - rightDistance;
            return new Point(x, y);
        }
    };

    /**
     * An array containing all the values of this enum. Used internally by the
     * randomDirection method.
     */
    private static final Direction[] values = values();

    /**
     * The direction considered left (90 degrees counterclockwise) of the given direction.
     * 
     * @return direction that is left
     */
    public abstract Direction left();

    /**
     * The direction considered right (90 degrees clockwise) of the given direction.
     * 
     * @return direction that is right
     */
    public abstract Direction right();

    /**
     * The direction considered back or behind (180 degrees) the given direction.
     * 
     * @return direction that is back
     */
    public abstract Direction back();

    /**
     * Returns the coordinate that ahead and to the right of a given coordinate,
     * ahead and right being relative to the direction faced. Either value can
     * be negative (being back and left, respectively). For example, the point 5
     * ahead and -3 right of the point (0, 0) when facing east is (5, 3)
     * 
     * @param point The initial coordinate.
     * @param foreDistance Distance in front of the point (or behind, if
     *        negative).
     * @param rightDistance Distance to the right of the point (or left, if
     *        negative).
     * @return The point.
     */
    public abstract Point getOffsetPoint(Point point, int foreDistance,
            int rightDistance);

    /**
     * Returns a random direction, using a given Random number generator. Used
     * to allow seeded games to turn out the same.
     * 
     * @param random The random number generator to use.
     * @return A randomly choosen Direction.
     */
    public static final Direction randomDirection(Random random) {
        int rNum = random.nextInt(4);
        return values[rNum];
    }
    
    /**
     * Returns this Direction as a String. Returns the direction converted to
     * lowercase.
     * 
     * @return The direction as a String.
     */
    @Override
    public String toString() {
        return name().toLowerCase();
    }

}
