package backend;

import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Formatter;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import exceptions.NotEnoughPlayersException;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * This class represents the "announcer" of the game or the gamekeeper. It
 * constructs the playing field, deals with the algorithms' moves , keeps score,
 * and keeps track of turns.
 * 
 * @author Seth Sweep
 */
public class GameLogic {

    /**
     * The version number of the game.
     */
    public static final String VERSION = "1.2";

    /**
     * The maximum number of turns in a game. If the playing field isn't
     * entirely captured by the time this many moves are up, the game is over
     * and the player with the most points wins.
     */
    private int _maxTurns = 5000;

    /**
     * The turn currently in progress.
     */
    private int _turn = 1;

    /**
     * The playing field where the cells (buildings and streets) are stored.
     */
    private PlayingField _playingField;

    /**
     * Set of Objects listening for changes in the game state.
     */
    private Set<GameListener> _listeners = new HashSet<GameListener>();

    /**
     * States if the game conclusion conditions have been satisifed (i.e.
     * whether the game is over).
     */
    private boolean _gameOver;

    /**
     * True if verbose logging is on, false otherwise.
     */
    private boolean _verbose;

    /**
     * The robots involved in the game.
     */
    private List<Robot> _robots = new ArrayList<Robot>();

    private int _delay;

    private boolean _throwExceptions;

    /**
     * The constructor for a new instance of the game logic with a given set of
     * algorithms, and a playing field determined by level. Currently, the three
     * levels supported are:
     * 
     * Level 1: 51 cells x 51 cells, no randomized buildings.
     * Level 2: 51 cells x 51 cells, randomized buildings.
     * Level 3: 61 cells x 81 cells, randomized buildings.
     * 
     * Any other value for level will return the same level as level 1. See
     * <code>PlayingField</code> for more information.
     * 
     * @param algorithms Array containing the algorithms for each robot.
     * @param level The level to use.
     * @throw NotEnoughPlayersException Thrown if there are no algorithms.
     * @throw TooManyPlayersException Thrown if there are too many algorithms.
     * @throw NullPointerException Thrown if the algorithms array is null, or
     *        any of the algorithms in the array are null.
     */
    public GameLogic(RobotAlgorithmIF[] algorithms, int level) {
        this(algorithms, level, null);
    }

    /**
     * Convenience constructor for a new instance of the game with a
     * non-variable playing field (level 1). This is the same as calling
     * <code>GameLogic(algorithms, 1)</code>.
     * 
     * @param algorithms Array containing the algorithms for each robot.
     * @throw NotEnoughPlayersException Thrown if there are no algorithms.
     * @throw TooManyPlayersException Thrown if there are too many algorithms.
     * @throw NullPointerException Thrown if the algorithms array is null, or
     *        any of the algorithms in the array are null.
     */
    public GameLogic(RobotAlgorithmIF[] algorithms) {
        this(algorithms, 1, null);
    }

    /**
     * The constructor for a new instance of the game logic with a given set of
     * algorithms, a playing field determined by level, and a given source of
     * randomness.
     * 
     * Two different instances of <code>GameLogic</code> using
     * <code>Random</code>s with the same seed are guaranteed to return the
     * same results, assuming the algorithms are deterministic.
     * 
     * Currently, the three levels supported are:
     * 
     * Level 1: 51 cells x 51 cells, no randomized buildings.
     * Level 2: 51 cells x 51 cells, randomized buildings.
     * Level 3: 61 cells x 81 cells, randomized buildings.
     * 
     * Any other value for level will return the same level as level 1. See
     * <code>PlayingField</code> for more information.
     * 
     * @param algorithms Array containing the algorithms for each robot.
     * @param level The level to use.
     * @param random The random number generator to use.
     * @throw NotEnoughPlayersException Thrown if there are no algorithms.
     * @throw TooManyPlayersException Thrown if there are too many algorithms.
     * @throw NullPointerException Thrown if the algorithms array is null, or
     *        any of the algorithms in the array are null.
     */
    public GameLogic(RobotAlgorithmIF[] algorithms, int level, Random random) {
        if (random == null)
            random = new Random();
        setup(algorithms, level, random);
    }

    /**
     * Initializes all the fields.
     * 
     * @param algorithms The algorithms to use.
     * @param level The level to use.
     * @param random The random source to use.
     * @throw NotEnoughPlayersException Thrown if there are no algorithms.
     * @throw TooManyPlayersException Thrown if there are too many algorithms.
     * @throw NullPointerException Thrown if the algorithms array is null, or
     *        any of the algorithms in the array are null.
     */
    private void setup(RobotAlgorithmIF[] algorithms, int level, Random random) {
        if (algorithms.length == 0) {
            throw new NotEnoughPlayersException("Illegal number of players: "
                    + algorithms.length);
        }
        if (Arrays.asList(algorithms).contains(null)) {
            throw new NullPointerException(
                    "The algorithms must all be non-null");
        }

        for (int i = 0; i < algorithms.length; i++) {
            Robot robot = new Robot(i + 1);
            robot.setAlgorithm(algorithms[i]);
            _robots.add(robot);
        }

        _playingField = new PlayingField(_robots, level, random);

        /*
         * If you wanted level-dependant numbers of moves, this is where you
         * would set that.
         */
    }

    /**
     * Runs the game. Iterates through the algorithms, setting the sensors and
     * getting moves. Each instance of GameLogic may be run only once, so will
     * throw an exception if it is called a second time.
     * 
     * @return Returns the scores of the robots, in the same order as they were
     *         given in the array in the constructor.
     * @throws IllegalStateException Thrown if this method has already been
     *         called.
     */
    public int[] execute() {
        if (_gameOver) {
            throw new IllegalStateException("The game has already been run");
        }
        notifyListeners();
        doGameTurns();
        _gameOver = true;
        notifyListeners();
        return getScores();
    }

    private void doGameTurns() {
        for (; _turn <= _maxTurns; _turn++) {
            long startTime = System.currentTimeMillis();
            for (Robot robot : _robots) {
                doOneRobotMove(robot);
            }
            long totalTime = System.currentTimeMillis() - startTime;
            if (totalTime < _delay) {
                sleep(_delay - totalTime);
            }
            if (_playingField.allBuildingsClaimed()) {
                log("All buildings claimed.");
                break;
            }
        }
        if (_turn == _maxTurns + 1) {
            _turn--;
        }
    }

    private void sleep(long time) {
        try {
            Thread.sleep(time);
        } catch (InterruptedException e) {
        }
    }

    private void doOneRobotMove(Robot robot) {
        RobotAlgorithmIF currentAlgorithm = robot.getAlgorithm();
        logTurn(robot);
        SensorGroup sensorData = robot.getSensors();

        Move move = null;

        try {
            move = currentAlgorithm.getMove(sensorData);
        } catch (RuntimeException e) {
            log(String.format("Player %s threw a %s", robot.getID(), e));
            e.printStackTrace(System.err);
            if (_throwExceptions) {
                throw e;
            } else {
                return;
            }
        }

        if (move == null) {
            log("Player " + robot.getID() + " returned a null move");
            return;
        }

        moveRobot(robot, move);
    }

    private void moveRobot(Robot robot, Move move) {
        logMove(move, robot);
        Point oldLocation = robot.getLocation();
        robot.checkAndMakeMove(move);
        Point newLocation = robot.getLocation();

        if (robot.crumbsCleared()) {
            Rectangle crumbArea = robot.crumbArea();
            notifyListeners(crumbArea.x, crumbArea.y, crumbArea.width,
                    crumbArea.height);
        } else {
            notifyListenersAboutRobotMove(oldLocation, newLocation);
        }
    }

    private void notifyListenersAboutRobotMove(Point oldLocation,
            Point newLocation) {
        final int oldX = oldLocation.x;
        final int newX = newLocation.x;
        final int oldY = oldLocation.y;
        final int newY = newLocation.y;
        notifyListeners(
                Math.min(oldX, newX), 
                Math.min(oldY, newY), 
                Math.max(oldX, newX), 
                Math.max(oldY, newY));
    }

    /**
     * Returns a multi-line formatted String consisting of each robot's score
     * and declaring the winner(s). If the game is still in progress, "Game in
     * progress." will be returned instead. A typical example of the output is:
     * 
     * <code>Game Over.
     * Game finished in 4438 turns!
     * ___End Game Report___
     * Robot 1 (MyRobotAlgorithm) claimed 157 points.
     * Robot 2 (ThisRobotAlgorithm) claimed 157 points.
     * Robot 3 (OtherRobotAlgorithm) claimed 82 points.
     * *** Robot 1 (MyRobotAlgorithm) Wins! ***
     * *** Robot 2 (ThisRobotAlgorithm) Wins! ***</code>
     */
    public String endGameReport() {
        if (!_gameOver) {
            return "Game in progress.";
        }
        StringBuilder result = new StringBuilder();
        Formatter formatter = new Formatter(result);
        result.append("Game Over.\n");
        if (getCurrentTurn() != getMaxTurns()) {
            formatter.format("Game finished in %s turns!\n", getCurrentTurn());
        }
        result.append("___End Game Report___\n");
        for (Robot robot : _robots) {
            formatter.format("Robot %s (%s) claimed %d points.\n", robot
                    .getID(), robot.getAlgorithm().getName(), robot.getScore());
        }
        for (Robot robot : getWinners()) {
            formatter.format("*** Robot %d (%s) Wins! ***\n", robot.getID(),
                    robot.getAlgorithm().getName());
        }
        return result.toString();
    }

    /**
     * Returns the robot(s) who are currently winning the game (have the most
     * points).
     * 
     * @return A Set of Robots consisting of the robot(s) winning.
     */
    private Set<Robot> getWinners() {
        int winningScore = Integer.MIN_VALUE;
        Set<Robot> result = new LinkedHashSet<Robot>();
        for (Robot robot : _robots) {
            int currentScore = robot.getScore();
            if (currentScore > winningScore) {
                result.clear();
            }
            if (currentScore >= winningScore) {
                result.add(robot);
                winningScore = currentScore;
            }
        }
        return result;
    }

    /**
     * Gives verbose text output of the current turn for the current robot.
     * 
     * @param robot The robot moving.
     */
    private void logTurn(Robot robot) {
        if (_verbose) {
            StringBuilder output = new StringBuilder();
            Formatter formatter = new Formatter(output);
            formatter.format("Move (%d/%d)", _turn, _maxTurns);
            formatter.format(" -- Player %d", robot.getID());
            formatter.format(" -- Score: %d", robot.getScore());
            formatter.format(" -- Robot Location: (%d, %d) facing %s.", robot
                    .getLocation().x, robot.getLocation().y, robot.getDirection());
            log(output.toString());
        }
    }

    /**
     * Gives verbose text output for a given player's move.
     * 
     * @param move The move being made.
     * @param robot The robot making the move.
     */
    private void logMove(Move move, Robot robot) {
        log(String.format("Player %s %s.", robot.getID(), move
                .getActionString()));
    }

    private void log(String message) {
        if (_verbose)
            System.out.println("Game Logic: " + message);
    }

    /**
     * Add the specified GameListener to the list of listeners to be notified
     * when the game state changes.
     * 
     * @param listener the listener to add.
     */
    public void addListener(GameListener listener) {
        _listeners.add(listener);
    }

    /**
     * Notify all the registered GameListeners that there's been some change to
     * the game state.
     */
    private void notifyListeners() {
        for (GameListener listener : _listeners) {
            listener.fireGameViewChanged();
        }
    }

    /**
     * Notify all the registered GameListeners that there's been some change to
     * the game state, and that the only changes to the playing field are in the
     * area in the given dimensions have changed on the playing field.
     * 
     * @param x The leftmost x coordinate of the changed area.
     * @param y The topmost y coordinate of the changed area.
     * @param width The width of the changed area.
     * @param height The height of the changed area.
     */
    private void notifyListeners(int x, int y, int width, int length) {
        for (GameListener listener : _listeners) {
            listener.fireGameViewChanged(x, y, width, length);
        }
    }

    /**
     * Sets whether the output should be verbose. If verbose is set to true,
     * various behind-the-scenes details will be logged to standard output.
     * Default value is false.
     * 
     * @param verbose Whether
     */
    public void setVerbose(boolean verbose) {
        _verbose = verbose;
        for (Robot robot : _robots) {
            robot.setVerbose(verbose);
        }
    }

    /**
     * Sets the minimum amount of time each turn should take, in miliseconds.
     * This allows you to slow down the game if it is progressing too quickly.
     * If the value is less than or equal to zero, each turn will finish as soon
     * as each robot has made its move.
     * 
     * @param delay The minimum time in miliseconds that a turn should take.
     */
    public void setTurnDelay(int delay) {
        _delay = delay;
    }

    /**
     * Set whether you want the game to catch exceptions thrown by any of the
     * algorithms. Used to prevent one bot to prematurely end the game, as may
     * be the case in most tournaments. Setting this to true is useful for
     * algorithm designers creating their bots, so they can find out where and
     * why exceptions are being thrown.
     * 
     * The default value is false.
     * 
     * @param throwExceptions Whether exceptions should be allowed to be thrown
     */
    public void setThrowBotExceptions(boolean throwExceptions) {
        _throwExceptions = throwExceptions;
    }

    /**
     * Returns the scores of the algorithms, in the same order as they were
     * given in the array in the constructor.
     * 
     * @return The scores of the algorithms.
     */
    public int[] getScores() {
        int[] result = new int[_robots.size()];
        for (int i = 0; i < result.length; i++) {
            result[i] = _robots.get(i).getScore();
        }
        return result;
    }

    /**
     * Returns the playing field used in this game.
     * 
     * @return The playing field.
     */
    public PlayingField getPlayingField() {
        return _playingField;
    }

    /**
     * Returns the maximum number of moves each robot gets in a game. If the
     * playing field isn't entirely captured by the time each robot has made
     * this many many moves, the game is over and the player with the most
     * points wins.
     * 
     * @return The maximum number of moves each robot gets.
     */
    public int getMaxTurns() {
        return _maxTurns;
    }

    /**
     * Returns the current turn number in the game.  A turn is defined as
     * each robot getting one move.
     * @return The current turn.
     */
    public int getCurrentTurn() {
        return _turn;
    }

    /**
     * Returns true if the game is over. This happens when either the maximum
     * number of turns has elapsed, or every building on the playing field has
     * been captured.
     * 
     * @return Whether the game is over.
     */
    public boolean isGameOver() {
        return _gameOver;
    }

    /**
     * Returns the level of the game, as defined in the constructor.
     * 
     * @return The level.
     */
    public int getLevel() {
        return _playingField.getLevel();
    }
}
