package backend;

import java.awt.Point;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import exceptions.MultipleRobotsInOnePlaceException;
import exceptions.NotEnoughPlayersException;
import exceptions.TooManyPlayersException;


/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * This class defines the playing field of the game. It consists of a grid of
 * cells, each representing a building or street.
 * 
 * Created by Seth Sweep
 * Heavily refactored by Matthew Justin, Matthew Carlson, and Jon Quarfoth.
 * 
 * @author Seth Sweep
 */
public class PlayingField implements BuildingCaptureListener {
    /**
     * The game level.
     */
    private int _level;

    /**
     * The collection of cells making up the field.
     */
    private Cell[][] _theField;

    /**
     * The field width. For the game to work properly, this should be an odd
     * number.
     */
    private int _fieldWidth = 51;

    /**
     * The field length. For the game to work properly, this should be an odd
     * number.
     */
    private int _fieldHeight = 51;

    /**
     * The random number generator used.
     */
    private final Random _random;

    /**
     * A list of the robots on this field.
     */
    private List<Robot> _robots = new ArrayList<Robot>();

    private int _unclaimedBuildings;

    /**
     * The coors of the level 2 "random building zone" (upper left coor, lower
     * right coor, and the number of buildings to try to place in that area)
     */
    private static final RandomZone L2_RANDOM_ZONE = 
            new RandomZone(10, 10, 40, 40, 50);

    /**
     * The coors of the level 3 "random building zones" (upper left coor, lower
     * right coor, and the number of buildings to try to place in that area)
     */
    private static final RandomZone[] L3_RANDOM_ZONES = {
            new RandomZone(1, 1, 36, 30, 50),
            new RandomZone(44, 1, 80, 30, 50),
            new RandomZone(1, 30, 36, 60, 50),
            new RandomZone(44, 30, 80, 60, 50), };

    /**
     * Convenience constructor. Identical to calling
     * <code>PlayinField(players, level, null)</code>.
     * 
     * @param players The number of players.
     * @param level The level.
     * @throws TooManyPlayersException If there are too many players to fit in
     *         the playing field.
     * @throws NotEnoughPlayersException If there are less than 1 player.
     */
    public PlayingField(int players, int level) throws TooManyPlayersException,
            NotEnoughPlayersException {
        this(players, level, new Random());
    }

    /**
     * Constructs a new PlayingField with the given number of players, a given
     * level, and a given random seed. The players created will have IDs of 1
     * through n, where n = the number of players. See the documentation of
     * PlayingField(List<Robot>, level, random) for more information.
     * 
     * @param players The number of players.
     * @param level The level.
     * @param random The random number generator to use, null to use a random
     *        number generator seeded by the system time.
     * @throws TooManyPlayersException If there are too many players to fit in
     *         the playing field.
     * @throws NotEnoughPlayersException If there are less than 1 player.
     */
    public PlayingField(int players, int level, Random random)
            throws TooManyPlayersException, NotEnoughPlayersException {
        this(defaultRobots(players), level, random);
    }
    
    private static List<Robot> defaultRobots(int players)
            throws NotEnoughPlayersException {
        List<Robot> result = new ArrayList<Robot>();
        for (int i = 0; i < players; i++) {
            result.add(new Robot(i + 1));
        }
        return result;
    }

    /**
     * The constructor for a new instance of the playing field with a given
     * group of players, a given random source, and a given level. Currently,
     * the three levels supported are:
     * 
     * Level 1: 51 cells x 51 cells, no randomized buildings.
     * Level 2: 51 cells x 51 cells, randomized buildings.
     * Level 3: 61 cells x 81 cells, randomized buildings.
     * 
     * Any other value for level will return the same level as level 1.
     * 
     * The default playing field for level 1 uses the following grid pattern:
     * 
     *    |sssssssssssssssss
     *  L |sBsBsBsBsBsBsBsBs
     *  E |sssssssssssssssss
     *  N |sBsBsBsBsBsBsBsBs
     *  G |sssssssssssssssss
     *  T |sBsBsBsBsBsBsBsBs
     *  H |sssssssssssssssss
     *    |-----------------
     *              WIDTH
     *              
     * B is a building.
     * s is a street.
     * 
     * At level 1, the buildings locations are not randomized, and they all have
     * point value 1.  Levels 2 & 3 randomize the point values of the buildings
     * (between 1 and 5), and add buildings randomly in the internal area of the
     * map.
     * 
     * Players 1 - 4 will be placed in the corners; players 5 - n are randomly
     * placed along the outside edge of the field, facing inwards.  If there
     * aren't enough spaces on the edge for each player, a
     * TooManyPlayersException will be thrown.
     * 
     * Two different playing fields using random number generators seeded to the
     * same value are guaranteed to return the same configuration of buildings
     * (values and locations) and player locations.
     * 
     * If random is null, uses a random number generator seeded by the system
     * time.
     * 
     * @param robots The robots to use in the game.
     * @param level The level type (1, 2, or 3); anything else defaults to 1.
     * @param random The random source to be used for this PlayingField.
     * @throws NotEnoughPlayersException If there are less than 1 player.
     * @throws TooManyPlayersException if there are too many players.
     */
    public PlayingField(List<Robot> robots, int level, Random random)
            throws TooManyPlayersException, NotEnoughPlayersException {
        if (robots == null) {
            throw new NullPointerException("robots must be non-null");
        }
        if (robots.size() == 0) {
            throw new NotEnoughPlayersException(
                    "There must be at least 1 robot");
        }
        if (random == null) {
            random = new Random();
        }
        _robots.addAll(robots);
        _level = level;
        _random = random;
        initializePlayingField(level);
        initializePlayers();
    }

    /**
     * Initializes the playing field, creating the cell grid and placing the
     * buildings.
     * @param level The level.
     * @throws TooManyPlayersException Thrown if there are too many players.
     */
    private void initializePlayingField(int level) {
        initializeFieldSize(level);

        if (tooManyPlayers(_robots.size())) {
            throw new TooManyPlayersException(_robots.size());
        }

        _theField = new Cell[_fieldWidth][_fieldHeight];

        setUpCellGrid(level);
    }

    private void initializeFieldSize(int level) {
        if (level == 1) {
            _fieldHeight = 51;
            _fieldWidth = 51;
        } else if (level == 2) {
            _fieldHeight = 51;
            _fieldWidth = 51;
        } else if (level == 3) {
            _fieldHeight = 61;
            _fieldWidth = 81;
        }
    }

    private void setUpCellGrid(int level) {
        boolean random = (level == 2 || level == 3);
        setUpBasicCellGrid(random);

        if (level == 2) {
            placeRandomBuildings(L2_RANDOM_ZONE);
        } else if (level == 3) {
            for (RandomZone zone : L3_RANDOM_ZONES) {
                placeRandomBuildings(zone);
            }
        }
    }

    private void placeRandomBuildings(RandomZone zone) {
        int buildings = zone.buildingsToPlace;

        while (buildings != 0) {
            int x = (_random.nextInt(zone.width)) + zone.x;
            int y = (_random.nextInt(zone.height)) + zone.y;
            Cell cell = getCell(x, y);
            if (cell instanceof StreetCell) {
                _theField[x][y] = createBuilding(_random.nextInt(5) + 1);
                buildings--;
            }
        }
    }

    private BuildingCell createBuilding(int value) {
        _unclaimedBuildings++;
        BuildingCell result = new BuildingCell(value);
        result.setListener(this);
        return result;
    }

    /**
     * Sets up the basic cell grid (but not the extra random buildings). If
     * random is true, each building is assigned a random value between 1 and 5,
     * otherwise the buiding's value is 1.
     * 
     * @param random Whether the point values of the buildings should be
     *        randomly determined
     */
    private void setUpBasicCellGrid(boolean random) {
        for (int y = 0; y < _fieldHeight; y++) {
            for (int x = 0; x < _fieldWidth; x++) {
                if (y % 2 == 1 && x % 2 == 1) {
                    _theField[x][y] = createBuilding(buildingValue(random));
                } else {
                    _theField[x][y] = new StreetCell();
                }
            }
        }
    }

    /**
     * Returns the value of a new building to place. If random is true, a random
     * value between 1 and 5 is returned, otherwise this returns 1.
     * 
     * @param random Whther the value should be randomly determined.
     * @return The value for the building.
     */
    private int buildingValue(boolean random) {
        if (random) {
            return _random.nextInt(5) + 1;
        } else {
            return 1;
        }
    }

    private boolean tooManyPlayers(int players) {
        int maxSlots = 2 * _fieldHeight + 2 * _fieldWidth - 4;
        return maxSlots < players;
    }

    /**
     * Sets up the initial player locations, as well as whatever other player
     * setup is necessary.
     */
    private void initializePlayers() {
        List<Robot> robotPositions = getStartingPositions(_robots.size());

        for (int i = 0; i < robotPositions.size(); i++) {
            Robot robotPosition = robotPositions.get(i);
            Robot robot = _robots.get(i);
            setRobotLocation(robot, robotPosition.getLocation());
            robot.setDirection(robotPosition.getDirection());
            robot.setSensor(new SensorGenerator(this, robot));
            robot.setReferee(new Referee(this, robot));
        }
    }

    private List<Robot> getStartingPositions(int players) {
        final List<Robot> result = new ArrayList<Robot>(Arrays.asList(
                new Robot(0, 0, Direction.EAST), new Robot(_fieldWidth - 1,
                        _fieldHeight - 1, Direction.WEST), new Robot(
                        _fieldWidth - 1, 0, Direction.SOUTH), new Robot(0,
                        _fieldHeight - 1, Direction.NORTH)));
        if (players <= 4) {
            return result.subList(0, players);
        } else {
            List<Robot> randomSlots = possibleRobotLocations();
            result.addAll(randomSlots.subList(0, players - 4));
            return result;
        } 
    }

    /**
     * Returns a randomized list consisting of robots in all possible starting
     * locations.
     * 
     * @return The randomized list.
     */
    private List<Robot> possibleRobotLocations() {
        List<Robot> result = new ArrayList<Robot>();
        for (int i = 1; i < _fieldWidth - 1; i++) {
            result.add(new Robot(i, _fieldHeight - 1, Direction.NORTH));
            result.add(new Robot(i, 0, Direction.SOUTH));
        }
        for (int i = 1; i < _fieldHeight - 1; i++) {
            result.add(new Robot(_fieldWidth - 1, i, Direction.WEST));
            result.add(new Robot(0, i, Direction.EAST));
        }
        Collections.shuffle(result, _random);
        return result;
    }

    /**
     * Sets the position and direction of a given robot in the field.
     * 
     * @param robot The robot to set the position of.
     * @param dir The robot's direction.
     * @param x The robot's x value.
     * @param y The robot's y value.
     * @throws MultipleRobotsInOnePlaceException Thrown if a different robot is
     *         already in the given cell
     */
    public void setRobot(Robot robot, Direction dir, int x, int y) {
        setRobotLocation(robot, new Point(x, y));
        robot.setDirection(dir);
    }

    /**
     * Sets the location of a given robot in the field.
     * 
     * @param robot
     * @param location
     */
    public void setRobotLocation(Robot robot, Point location) {
        Cell cell = getCell(location);
        if (!(cell instanceof StreetCell)) {
            throw new IllegalStateException(
                    "Attempted to place a robot on a non-Street cell");
        }
        StreetCell newCell = (StreetCell) cell;
        StreetCell oldCell = (StreetCell) getCell(robot.getLocation());

        oldCell.setRobotOccupying(robot, false);
        newCell.setRobotOccupying(robot, true);
        robot.setLocation(location);
    }

    /**
     * Returns the robots on this field.
     * 
     * @return A list containing the robots.
     */
    public List<Robot> getRobots() {
        return Collections.unmodifiableList(_robots);
    }

    /**
     * Gets the cell at the given coordinates. If the coordinates are out of
     * bounds, returns the singleton instance of WallCell.
     * 
     * @param point The coordinates to get the cell at.
     * @return The cell at that location.
     */
    public Cell getCell(Point point) {
        return getCell(point.x, point.y);
    }

    /**
     * Gets the cell at the given coordinates. If the coordinates are out of
     * bounds, returns the singleton instance of WallCell.
     * 
     * @param x The x coordinate.
     * @param y The y coordinate.
     * @return The cell at that location.
     */
    public Cell getCell(int x, int y) {
        if (cellOutOfBounds(x, y)) {
            return WallCell.getInstance();
        } else {
            return _theField[x][y];
        }
    }

    private boolean cellOutOfBounds(int x, int y) {
        return x < 0 || x >= getWidth() || y < 0
                || y >= getHeight();
    }

   /**
	 * Returns the width of the playing field.
	 * 
	 * @return The width of the playing field.
	 */
	public int getWidth() {
        return _fieldWidth;
    }

    /**
     * Returns The height of the playing field.
     * 
     * @return The height of the playing field.
     */
    public int getHeight() {
        return _fieldHeight;
    }

    /**
     * Goes through and cleares the checks of all cells.
     */
    public void clearChecked() {
        for (int i = 0; i < _fieldWidth; i++) {
            for (int j = 0; j < _fieldHeight; j++) {
                Cell cell = getCell(i, j);
                cell.setChecked(false);
            }
        }
    }

    /**
     * Returns true if all buildings on the field have been claimed.
     * 
     * @return true iff all buildings on the field have been claimed.
     */
    public boolean allBuildingsClaimed() {
        return _unclaimedBuildings == 0;
    }

    /**
     * Returns the field level.
     * @return The level of the field.
     */
    public int getLevel() {
        return _level;
    }
    
    public void fireBuildingChanged(BuildingCell cell) {
        _unclaimedBuildings--;
    }

    /**
     * Represents a random zone, used to place buildings in levels 2 & 3.  The
     * program places <code>buildingsToPlace</code> number of buildings in the
     * area with upper-left corner (x, y) with determined width & height.
     * 
     * @author Matthew Justin
     */
    private static class RandomZone {
        public int x;

        public int y;

        public int width;

        public int height;

        public int buildingsToPlace;

        public RandomZone(int left, int top, int right, int bottom,
                int buildingsToPlace) {
            x = left;
            y = top;
            width = right - left;
            height = bottom - top;
            this.buildingsToPlace = buildingsToPlace;
        }
    }
}
