package backend;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * Object which scans the playing field to determine whether a legal capture has
 * been made for a given robot.
 * 
 * @author Seth Sweep
 */
public class RecursiveEnclosureScanner {

    private PlayingField _playingField;
    private Robot _robot;
    private List<Cell> _cellsChecked = new ArrayList<Cell>();

    /**
     * Creates a new scanner.
     * 
     * @param field The PlayingField to scan.
     * @param robot The Robot to scan for captures of.
     */
    public RecursiveEnclosureScanner(PlayingField field, Robot robot) {
        _playingField = field;
        _robot = robot;
    }

    /**
     * Scans the playing field for the given robot's crumbs, based on the
     * robot's current position. Returns the number of points that should be
     * awarded to the robot.
     * 
     * @return The points that should be awarded.
     */
    public int awardPoints() {
        Point startPoint = findPointInsideValidEnclosure();
        if (startPoint == null) {
            return 0;
        }
        int result = recursivelyAwardPoints(startPoint);
        clearChecked();
        return result;
    }

    private void clearChecked() {
        for (Cell cell : _cellsChecked) {
            cell.setChecked(false);
        }
        _cellsChecked.clear();
    }

    private void setChecked(Cell cell) {
        cell.setChecked(true);
        _cellsChecked.add(cell);
    }

    /**
     * Attempts to find a point inside the enclosure. Returns null if no point
     * is inside the enclosure, or if the enclosure contains buildings belonging
     * to other robots.
     * 
     * @return A point inside the enclosure, or null if no such point exists.
     */
    private Point findPointInsideValidEnclosure() {
        for (int ahead = -1; ahead <= 1; ahead++) {
            for (int right = -1; right <= 1; right++) {
                Point location = offsetPoint(_robot.getLocation(), ahead, right);
                if (isInsideValidEnclosure(location)) {
                    clearChecked();
                    return location;
                }
                clearChecked();
            }
        }
        clearChecked();
        return null;
    }

    /**
     * Returns a copy of the point passed in, offset by xOffset and yOffset.
     * 
     * @param point The point to return an offset copy of.
     * @param xOffset The x offset.
     * @param yOffset The y offset.
     * @return An offset copy of the point.
     */
    private Point offsetPoint(Point point, int xOffset, int yOffset) {
        Point result = new Point(point);
        result.translate(xOffset, yOffset);
        return result;
    }

    private boolean isInsideValidEnclosure(Point location) {
        Cell cell = _playingField.getCell(location);
        if (cell.isChecked() || streetWithRobotsTrail(cell)) {
            return false;
        }
        return isValidNeighbor(location);
    }

    private boolean streetWithRobotsTrail(Cell cell) {
        return cell instanceof StreetCell
                && ((StreetCell) cell).hasRobotTrail(_robot);
    }

    /**
     * Returns true if this location contains a valid neighboring cell for a
     * point within a legal enclosure.
     * 
     * @param location The cell location.
     * @return True iff this is a valid neighboring cell.
     */
    private boolean isValidNeighbor(Point location) {
        Cell cell = _playingField.getCell(location);

        if (cell instanceof WallCell) {
            return false;
        }

        if (cell.isChecked()) {
            return true;
        } else {
            setChecked(cell);
        }

        if (cell instanceof BuildingCell) {
            return isValidBuildingNeighbor(location, (BuildingCell) cell);
        } else {
            return isValidStreetNeighbor(location, (StreetCell) cell);
        }
    }

    private boolean isValidStreetNeighbor(Point location,
            StreetCell sCell) {
        if (sCell.hasRobotTrail(_robot)) {
            return true;
        } else {
            return areNeighborsValid(location);
        }
    }

    private boolean isValidBuildingNeighbor(Point location,
            BuildingCell bCell) {
        if (bCell.isOwned() && bCell.getOwner() != _robot) {
            return false;
        } else {
            return areNeighborsValid(location);
        }
    }

    private boolean areNeighborsValid(Point location) {
        boolean valid = true;
        for (int x = -1; x <= 1; x++) {
            for (int y = -1; y <= 1; y++) {
                if (x == 0 && y == 0) {
                    continue;
                }
                valid = valid
                        && isValidNeighbor(offsetPoint(location, x, y));
            }
        }
        return valid;
    }

    /**
     * Awards points by checking the value of the cell in the given location
     * (and setting the owner, if appropriate), and then checking the value of
     * all the cell's neighbors. If the cell was already checked, or if the cell
     * contains the robot's crumbs, returns 0.
     * 
     * @param location The location to check at.
     * @return The points to be awarded.
     */
    private int recursivelyAwardPoints(Point location) {
        Cell cell = _playingField.getCell(location);
        if (cell.isChecked() || streetWithRobotsTrail(cell)) {
            return 0;
        }

        setChecked(cell);

        int result = 0;
        if (cell instanceof BuildingCell) {
            BuildingCell bCell = (BuildingCell) cell;
            if (!bCell.isOwned()) {
                bCell.setOwner(_robot);
                result = bCell.getPoints();
            }
        }
        return result + neighborsPointValues(location);
    }

    /**
     * Returns the point values of all unchecked neighbors (plus the values of
     * their neighbors, etc.).
     * 
     * @param location The location to check the neighbors of.
     * @return The point values of the neighbors.
     */
    private int neighborsPointValues(Point location) {
        int result = 0;
        for (int x = -1; x <= 1; x++) {
            for (int y = -1; y <= 1; y++) {
                if (x == 0 && y == 0) {
                    continue;
                }
                result += recursivelyAwardPoints(offsetPoint(location,
                        x, y));
            }
        }
        return result;
    }
}
