package backend;

import java.awt.Point;
import java.awt.Rectangle;
import java.util.ArrayList;
import java.util.List;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * This class defines the "referee" of the game. It could also be thought of as
 * the "trail checker." It scans the field checking for cell enclosures and
 * awards points to valid captures. It also removes trails after valid captures,
 * if a robot chooses to "lift the pen", or if the robot attempts to double back
 * upon itself
 * 
 * @author Seth Sweep
 */
public class Referee {
    private PlayingField _playingField;

    private Robot _robot;

    private List<Point> _trail = new ArrayList<Point>();

    private RecursiveEnclosureScanner _scanner;

    /**
     * Used for determining if the view should update.
     */
    private boolean _crumbsCleared;

    private boolean _verbose;

    private Rectangle _crumbArea;

    /**
     * Creates a new referee with a given PlayingField and a given Robot.
     * 
     * @param playingField The PlayingField to use.
     * @param robot The robot to use.
     */
    public Referee(PlayingField playingField, Robot robot) {
        _playingField = playingField;
        _robot = robot;

        _scanner = new RecursiveEnclosureScanner(_playingField, robot);
    }

    /**
     * Lays down crumbs to the cell behind the robot if appropriate. Awards
     * points to the robot if a successful capture is made. It also will remove
     * crumbs if a robot completes an enclosure, moves ahead without crumbs, or
     * moves backwards onto its own trail.
     * 
     * @param move The move made by the robot.
     * @return The points earned from this move.
     */
    public int checkAndMakeMove(Move move) {
        if (move != Move.AHEAD_WITH_CRUMBS && move != Move.AHEAD_WITHOUT_CRUMBS) {
            return 0;
        }
        Point oldLocation = _robot.getLocation();
        Point newLocation = _robot.getDirection().getOffsetPoint(oldLocation, 1, 0);
        if (locationValidToMoveTo(newLocation)) {
            _playingField.setRobotLocation(_robot, newLocation);
            return crumbManager(move, oldLocation);
        } else {
            return 0;
        }
    }

    /**
     * Lays down crumbs to the cell behind the robot if appropriate. Awards
     * points to the robot if a successful capture is made. It also will remove
     * crumbs if a robot completes an enclosure, moves ahead without crumbs, or
     * moves backwards onto its own trail.
     * 
     * @param move The move made by the robot.
     * @param location The location the robot is moving from.
     * @return The points earned from this move.
     */
    private int crumbManager(Move move, Point location) {
        if (move == Move.AHEAD_WITH_CRUMBS) {
            return moveWithCrumbsAndGetScore(new Point(location));
        } else if (move == Move.AHEAD_WITHOUT_CRUMBS) {
            if (!_trail.isEmpty()) {
                clearCrumbs();
            }
        }
        return 0;
    }
    
    /**
     * Insures that a cell at coordinates (x,y) is okay for a robot to move to.
     * A cell is okay for moving to if a cell is not a wall (out of bounds), is
     * not a building, and is not currently occupied by another robot.
     * 
     * @param cellLocation The location of the cell a robot is trying to move to.
     * @return true Iff cell is okay to move to.
     */
    private boolean locationValidToMoveTo(Point cellLocation) {
        Cell cell = _playingField.getCell(cellLocation);
        if (cell.isObstacle()) {
            log(cell.getObstacleMessage());
            return false;
        } else {
            return true;
        }
    }

    private int moveWithCrumbsAndGetScore(Point location) {
        StreetCell sCell = (StreetCell) _playingField.getCell(location);
        sCell.setRobotTrail(_robot, true);
        if (isMovingBackwards()) {
            clearCrumbs();
        }
        _trail.add(location);
        if (enclosureMade()) {
            clearExcessTrails();
            return awardPointsAndClearCrumbs();
        } else {
            return 0;
        }
    }

    /**
     * Clears all the trails prior to the start of the enclosure. Necessary to
     * prevent a bug where a figure-8 shaped enclosure would award the wrong
     * trails.
     */
    private void clearExcessTrails() {
        int crossing = _trail.lastIndexOf(_robot.getLocation());
        for (Point point : _trail.subList(0, crossing)) {
            StreetCell cell = (StreetCell) _playingField.getCell(point);
            cell.setRobotTrail(_robot, false);
        }
    }

    private Rectangle calculateCrumbArea() {
        int bottom = Integer.MAX_VALUE;
        int top = Integer.MIN_VALUE;
        int left = Integer.MAX_VALUE;
        int right = Integer.MIN_VALUE;

        for (Point point : _trail) {
            bottom = Math.min(bottom, point.y);
            top = Math.max(top, point.y);
            left = Math.min(left, point.x);
            right = Math.max(right, point.x);
        }
        Rectangle result = new Rectangle(left, bottom, right, top);
        return result;
    }

    /**
     * Returns true if any crumbs have been cleared since the last time this
     * method was called.
     * 
     * @return Whether an enclosure was made.
     */
    public boolean crumbsCleared() {
        boolean result = _crumbsCleared;
        _crumbsCleared = false;
        return result;
    }

    /**
     * Returns the bounding rectangle of the trail of crumbs most recently
     * cleared.
     * 
     * @return The bounding rectangle.
     */
    public Rectangle crumbArea() {
        return _crumbArea;
    }

    /**
     * Turns verbose logging on or off.
     * 
     * @param verbose Whether to use verbose logging.
     */
    public void setVerbose(boolean verbose) {
        _verbose = verbose;
    }

    /**
     * This checks if a robot is trying to move backwards on its own
     * breadcrumbs.
     * 
     * @return Whether the robot is moving backwards.
     */
    private boolean isMovingBackwards() {
        if (_trail.isEmpty()) {
            return false;
        }
        Point lastCoor = _trail.get(_trail.size() - 1);
        if (_robot.getLocation().equals(lastCoor)) {
            log("Robot " + _robot.getID()
                    + " is attempting to move backwards with crumbs,"
                    + " which is an illegal move.  Bird penalty.");
            return true;
        }
        return false;
    }

    private boolean enclosureMade() {
        if (_trail.contains(_robot.getLocation())) {
            log("Enclosure made!  "
                    + "Checking if this is a legal enclosure...");
            return true;
        }
        return false;
    }

    private int awardPointsAndClearCrumbs() {
        int points = _scanner.awardPoints();
        clearCrumbs();
        return points;
    }

    private void clearCrumbs() {
        _crumbArea = calculateCrumbArea();
        for (Point coor : _trail) {
            StreetCell sCell = (StreetCell) _playingField.getCell(coor);
            sCell.setRobotTrail(_robot, false);
        }
        _trail.clear();
            log("*A Flock of Birds* ate robot " + _robot.getID()
                + "'s birdcrumbs!");
        _crumbsCleared = true;
    }

    private void log(String message) {
        if (_verbose)
            System.out.println("Referee: " + message);
    }
}
