package backend;

import java.awt.Point;
import java.awt.Rectangle;

public class Robot {
    private Point _location;

    private Direction _direction;

    private int _ID;

    private int _score;

    private RobotAlgorithmIF _algorithm;

    private SensorGenerator _sensor;

    private Referee _referee;

    /**
     * Creates a new robot with a given ID at coordinates (0,0) and facing
     * north.
     * 
     * @param ID The robot's ID.
     */
    public Robot(int ID) {
        this(0, 0, Direction.NORTH);
        _ID = ID;
    }

    /**
     * Creates a new robot at the given coordinates facing a given direction and
     * ID 0.
     * 
     * @param x The x coordinate.
     * @param y The y coordinate.
     * @param direction The direction the robot is facing.
     */
    public Robot(int x, int y, Direction direction) {
        if (direction == null) {
            throw new NullPointerException("Arguments must be non-null.");
        }
        _location = new Point(x, y);
        _direction = direction;
    }

    /**
     * Returns the direction the robot is facing.
     * 
     * @return The direction the robot is facing.
     */
    public Direction getDirection() {
        return _direction;
    }

    /**
     * Sets the direction the robot is facing.
     * 
     * @param direction The direction the robot is facing.
     */
    public void setDirection(Direction direction) {
        _direction = direction;
    }

    /**
     * Returns the robot's location.
     * 
     * @return The location of the robot.
     */
    public Point getLocation() {
        return new Point(_location);
    }

    /**
     * Sets the robot's location.
     * 
     * @param location The location to set.
     */
    public void setLocation(Point location) {
        _location = new Point(location);
    }

    /**
     * Returns the robot's ID.
     * 
     * @return The robot's ID.
     */
    public int getID() {
        return _ID;
    }

    /**
     * Sets the robot's ID.
     * 
     * @param ID The ID to set.
     */
    public void setID(int ID) {
        _ID = ID;
    }

    /**
     * Returns the robot's score.
     * 
     * @return The score of the robot.
     */
    public int getScore() {
        return _score;
    }

    /**
     * Sets the algorithm for the robot to use.
     * 
     * @param algorithm The algorithm for the robot to use.
     */
    public void setAlgorithm(RobotAlgorithmIF algorithm) {
        _algorithm = algorithm;
    }

    /**
     * Returns the algorithm being used by the robot.
     * 
     * @return The algorithm being used by the robot.
     */
    public RobotAlgorithmIF getAlgorithm() {
        return _algorithm;
    }

    /**
     * Sets the robot's sensor generator.
     * 
     * @param sensor The generator to set.
     */
    public void setSensor(SensorGenerator sensor) {
        _sensor = sensor;
    }

    /**
     * Returns the sensor data of the robot given its current position.
     * 
     * @return The robot's sensor data.
     */
    public SensorGroup getSensors() {
        return _sensor.getSensors();
    }
    
    /**
     * Sets the robot's referee.
     * 
     * @param referee The referee to set.
     */
    public void setReferee(Referee referee) {
        _referee = referee;
    }

    /**
     * This method checks if the robot may make a particular move (based on its
     * referee), and if the move is indeed valid, does so.
     * 
     * @param move The move to be made
     * @throw NullPointerException Thrown if either argument is null.
     */
    public void checkAndMakeMove(Move move) {
        switch (move) {
        case SKIP: {
            return;
        }
        case TURN_LEFT: {
            _direction = _direction.left();
            return;
        }
        case TURN_RIGHT: {
            _direction = _direction.right();
            return;
        }
        case AHEAD_WITHOUT_CRUMBS:
        case AHEAD_WITH_CRUMBS: {
            _score += _referee.checkAndMakeMove(move);
            return;
        }
        default: // move is null
            throw new NullPointerException("Move was null");
        }
    }

    /**
     * Returns true if any crumbs have been cleared since the last time this
     * method was called.
     * 
     * @return Whether an enclosure was made.
     */
    public boolean crumbsCleared() {
        return _referee.crumbsCleared();
    }

    /**
     * Returns the bounding rectangle of the trail of crumbs most recently
     * cleared.
     * 
     * @return The bounding rectangle.
     */
    public Rectangle crumbArea() {
        return _referee.crumbArea();
    }

    /**
     * Turns verbose logging on or off.
     * 
     * @param verbose Whether to use verbose logging.
     */
    public void setVerbose(boolean verbose) {
        _referee.setVerbose(verbose);
    }
}
