package backend;

/**
 * This class represents the sensor data for a single cell.
 * 
 * @author Matthew Justin
 */
public class Sensor {
    /**
     * Enum representing the type of a cell (building, street, or wall).
     * 
     * @author justnm
     */
    public enum CellType {BUILDING, STREET, WALL};

    /**
     * Enum representing the various possible owners of a building (unclaimed,
     * owned by your robot, or owned by an adversary).
     * 
     * @author justnm
     */
    public enum BuildingOwner {
        UNCLAIMED("an unclaimed building"), 
        ME("a building belonging to me"), 
        ADVERSARY("a building belonging to my adversary");
        
        private String string;
        
        private BuildingOwner(String string) {
            this.string = string;
        }
        
        public String toString() {
            return string;
        }
    };

    private CellType cellType;
    private BuildingOwner buildingOwner = BuildingOwner.UNCLAIMED;
    private int points;
    private boolean hasMyCrumbs;
    private boolean hasAdversaryCrumbs;
    private boolean hasAdversary;

    /**
     * Creates a new Sensor for a given cell and a given Robot.
     * 
     * @param cell The cell to get sensor data for.
     * @param robot The player receiving the sensor data.
     */
    public Sensor(Cell cell, Robot robot) {
        cellType = cell.cellType();
        if (cell instanceof BuildingCell) {
            setUpBuildingCell((BuildingCell)cell, robot);
        } else if (cell instanceof StreetCell) {
            setUpStreetCell((StreetCell)cell, robot);
        }
    }

    /**
     * Returns the owner of the building. Possible values are:
     * 
     * UNCLAIMED (the building is unclaimed)
     * ME (the building is owned by the robot)
     * ADVERSARY (the building is owned by one of the robot's adversaries).
     * 
     * If this cell is not a building cell, returns UNCLAIMED.
     * 
     * @return The owner of the building, or UNCLAIMED if the cell is either an
     *         unclaimed building or unowned.
     */
    public BuildingOwner getOwner() {
        return buildingOwner;
    }

    /**
     * Returns the type of the cell. Possible values are BUILDING, STREET, and
     * WALL.
     * 
     * @return The type of the cell.
     */
    public CellType getCellType() {
        return cellType;
    }

    /**
     * Returns true if the cell contains one of the robot's adversaries.
     * 
     * @return Whether the cell contains one of the robot's adversaries.
     */
    public boolean hasAdversary() {
        return hasAdversary;
    }

    /**
     * Returns true if the cell contains breadcrumbs belonging to one of the
     * robot's adversaries.
     * 
     * @return Whether the cell contains breadcrumbs belonging to one of the
     *         robot's adversaries.
     */
    public boolean hasAdversarysCrumbs() {
        return hasAdversaryCrumbs;
    }

    /**
     * Returns true if the cell contains breadcrumbs belonging to the robot.
     * 
     * @return Whether the cell contains breadcrumbs belonging to the robot.
     */
    public boolean hasMyCrumbs() {
        return hasMyCrumbs;
    }

    /**
     * Returns the point value for capturing the cell. For non-building cells,
     * this returns 0. For building cells, returns the value of the building.
     * 
     * @return The point value of the cell.
     */
    public int getPointValue() {
        return points;
    }

    /**
     * Returns a String representation of the sensor in human readable form.
     */
    public String toString() {
        StringBuilder result = new StringBuilder();
        result.append("I detect ");
        if (cellType == CellType.WALL) {
            result.append("a wall ");
        } else if (cellType == CellType.BUILDING) {
            result.append(buildingOwner.toString());
            result.append(" worth " + points + " point(s) ");
        } else {
            result.append(getStreetString());
        }

        return result.toString();
    }

    private String getStreetString() {
        StringBuilder result = new StringBuilder();
        if (hasMyCrumbs) {
            result.append("a road with my breadcrumbs ");
            if (hasAdversaryCrumbs) {
                result.append("and my adversary's breadcrumbs ");
            }
        } else if (hasAdversaryCrumbs) {
            result.append("a road with my adversary's breadcrumbs ");
        } else {
            result.append("a road with no breadcrumbs ");
        }

        if (hasAdversary) {
            result.append("and my adversary ");
        }
        return result.toString();
    }

    private void setUpStreetCell(StreetCell cell, Robot robot) {
        hasMyCrumbs = cell.hasRobotTrail(robot);
        hasAdversaryCrumbs = cell.isCrumbyExceptFor(robot);
        hasAdversary = cell.isOccupied() && cell.getRobotOccupying() != robot;
    }

    private void setUpBuildingCell(BuildingCell cell, Robot robot) {
        points = cell.getPoints();
        if (!cell.isOwned()) {
           buildingOwner = BuildingOwner.UNCLAIMED;
        } else if (cell.getOwner() == robot) {
            buildingOwner = BuildingOwner.ME;
        } else {
            buildingOwner = BuildingOwner.ADVERSARY;
        }
    }
}
