package backend;

import java.awt.Point;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * This class contains a robot's sensors for a given turn.
 * 
 * If the PlayingField changes after a SensorGroup is created, previously
 * created SensorGroups will not reflect the change.  This allows SensorGroup
 * objects given to algorithms to be stored, if desired, to keep track of what
 * the game state was at a given turn.
 * 
 * @author Matthew Justin
 */
public class SensorGroup {
    private Sensor center;

    private Sensor left;
    private Sensor aheadLeft;
    private Sensor aheadOne;
    private Sensor aheadRight;
    private Sensor right;
    private Sensor farAhead;
    private int farSensorDistance;

    private PlayingField _playingField;
    private Robot _robot;

    /**
     * Creates a new sensor for the given robot in a given playing field.
     * 
     * @param playingField The playing field to use.
     * @param robot The robot to use.
     */
    public SensorGroup(PlayingField playingField, Robot robot) {
        _playingField = playingField;
        _robot = robot;

        center = sensorSquareWithOffset(0, 0);
        left = sensorSquareWithOffset(0, -1);
        aheadLeft = sensorSquareWithOffset(1, -1);
        aheadOne = sensorSquareWithOffset(1, 0);
        aheadRight = sensorSquareWithOffset(1, 1);
        right = sensorSquareWithOffset(0, 1);
        farAhead = sensorSquareWithOffset(distanceToObstacle(), 0);
        farSensorDistance = distanceToObstacle() - 1;
    }

    /**
     * Sensor data for the cell the robot is currently on.
     * 
     * @return The sensor reading for the cell.
     */
    public Sensor center() {
        return center;
    }

    /**
     * Sensor data for the cell to the left of the robot.
     * 
     * @return The sensor reading for the cell.
     */
    public Sensor left() {
        return left;
    }

    /**
     * Sensor data for the cell to the right of the robot.
     * 
     * @return The sensor reading for the cell.
     */
    public Sensor right() {
        return right;
    }

    /**
     * Sensor data for the cell ahead one and to the left one of the robot.
     * 
     * @return The sensor reading for the cell.
     */
    public Sensor aheadLeft() {
        return aheadLeft;
    }

    /**
     * Sensor data for the cell one ahead of the robot.
     * 
     * @return The sensor reading for the cell.
     */
    public Sensor aheadOne() {
        return aheadOne;
    }

    /**
     * Sensor data for the cell ahead one and to the right one of the robot.
     * 
     * @return The sensor reading for the cell.
     */
    public Sensor aheadRight() {
        return aheadRight;
    }

    /**
     * Sensor data for the first cell straight ahead of the robot that is an
     * obstacle.
     * 
     * @return The sensor reading for the cell.
     */
    public Sensor farAhead() {
        return farAhead;
    }

    /**
     * The distance ahead of the robot of the cell returned by the far ahead
     * sensor.
     * 
     * @return The distance to the nearest obstacle straight ahead of the robot.
     */
    public int farSensorDistance() {
        return farSensorDistance;
    }

    /**
     * Returns the sensor group as a seven element String array. Each position
     * of the array corresponds to a different sensor. The seven sensors (in the
     * order they appear in the array) are:
     * 
     * Center, Left, Ahead-left, Ahead-one, Ahead-right, Right, and Far Ahead.
     * 
     * Each String consists of the sensor name, followed by the sensor input in
     * the following format:
     * 
     * SensorName: SensorData
     * 
     * An example of sensor input is:
     * 
     * Ahead-right: I detect an unclaimed building worth 1 point(s)
     * 
     * If the sensor is the far sensor, the String will also display the
     * distance, as in this example:
     * 
     * Ahead: I detect a wall 50 spaces away
     * 
     * @return The sensor data, as a String array.
     */
    public String[] toStringArray() {
        String[] result = new String[7];

        result[0] = "Center: " + center;
        result[1] = "Left: " + left;
        result[2] = "Ahead-left: " + aheadLeft;
        result[3] = "Ahead-one: " + aheadOne;
        result[4] = "Ahead-right: " + aheadRight;
        result[5] = "Right: " + right;
        result[6] = "Ahead: " + farAhead + farSensorDistance + " spaces away ";

        return result;
    }

    /**
     * Returns the SensorGroup, as a String. Displays each sensor on a new line,
     * with an appropriate label for each. The output is formatted as:
     * 
     * Center: center() Left: left() Ahead-left: aheadLeft() Ahead-one:
     * aheadOne() Ahead-right: aheadRight() Right: right() Ahead: farAhead()
     * farSensorDistance() spaces away
     */
    public String toString() {
        StringBuilder result = new StringBuilder();
        for (String s : toStringArray()) {
            result.append(s + "\n");
        }
        result.deleteCharAt(result.length() - 1);
        return result.toString();
    }

    /**
     * Returns sensor data for the square offset from the robot's position by
     * the given amounts.
     * 
     * @param xOffset How far ahead of the robot the point is.
     * @param yOffset How far behind the robot the point is.
     * @return
     */
    private Sensor sensorSquareWithOffset(int xOffset, int yOffset) {
        Point cellLocation = _robot.getDirection().getOffsetPoint(
                _robot.getLocation(), xOffset, yOffset);
        return new Sensor(_playingField.getCell(cellLocation), _robot);
    }

    /**
     * Returns the distance to the nearest obstacle directly in front of the
     * robot.
     * 
     * @return The distance to the nearest obstacle.
     */
    private int distanceToObstacle() {
        int result = 1;
        while (!isObstacleAtDistance(result)) {
            result++;
        }
        return result;
    }

    /**
     * Returns true if the cell <code>distance</code> spaces ahead of the
     * robot is an obstacle.
     * 
     * @param distance The distance.
     * @return Whether the cell is an obstacle.
     */
    private boolean isObstacleAtDistance(int distance) {
        Point location = _robot.getDirection().getOffsetPoint(
                _robot.getLocation(), distance, 0);
        Cell cell = _playingField.getCell(location);
        return cell.isObstacle();
    }
}
