package backend;

import java.util.HashSet;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import exceptions.MultipleRobotsInOnePlaceException;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * This class defines a street cell. Each street keeps track of which robots
 * have left breadcrumbs in it, which robot (if any) is currently occupying it,
 * and a flag indicating whether this street was "checked" (used by the
 * enclosure scanning algorithm for awarding points).
 * 
 * @author Seth Sweep
 */
public class StreetCell implements Cell {

    /**
     * Set containing the Robots that have left trails in this cell.
     */
    private Set<Robot> _trails = new HashSet<Robot>();

    /**
     * Which robot occupies this cell.  Null if unoccupied.
     */
    private Robot _robotOccupying;

    /**
     * Has this been checked? Used by the enclosure scanning algorithm for
     * awarding points.
     */
    private boolean _checked = false;

    /**
     * Sets the trail of a robot in this cell.
     * 
     * @param robot The robot to set the trail of.
     * @param trail True if the robot has a trail in this cell, false otherwise.
     */
    public void setRobotTrail(Robot robot, boolean trail) {
        if (trail) {
            _trails.add(robot);
        } else {
            _trails.remove(robot);
        }
    }

    /**
     * Returns true if and only if trail of a given robot is in this cell.
     * 
     * @return Whether a given robot is in this cell.
     */
    public boolean hasRobotTrail(Robot robot) {
        return _trails.contains(robot);
    }

    /**
     * Sets whether a given robot occupies this cell. Throws a
     * MultipleRobotsInOnePlaceException if <code>occupies</code> is true and
     * a different robot already occupies this cell.
     * 
     * @param robot The Robot to add/remove from this cell.
     * @param occupies True if the robot should this cell, false otherwise
     * 
     * @throws MultipleRobotsInOnePlaceException Thrown if a different robot
     *         already occupies this cell and <code>occupies</code> is true..
     */
    public void setRobotOccupying(Robot robot, boolean occupies) {
        if (occupies) {
            if (anotherRobotOccupies(robot)) {
                throw new MultipleRobotsInOnePlaceException("Robot "
                        + _robotOccupying + " already occupies this space");
            }
            _robotOccupying = robot;
        } else {
            if (_robotOccupying == robot) {
                _robotOccupying = null;
            }
        }
    }

    private boolean anotherRobotOccupies(Robot robot) {
        return isOccupied() && _robotOccupying != robot;
    }

    /**
     * Returns the robot occupying this cell, or null if this cell is unocupied.
     * 
     * @return The robot occupying this cell.
     */
    public Robot getRobotOccupying() {
        return _robotOccupying;
    }

    /**
     * Returns true if this cell is occupied by a robot.
     * 
     * @return Whether this cell is occupied.
     */
    public boolean isOccupied() {
        return getRobotOccupying() != null;
    }

    /**
     * Returns true iff any breadcrumbs are on this cell.
     * 
     * @return True iff there are any breadcrumbs in this cell.
     */
    public boolean isCrumby() {
        return !(_trails.isEmpty());
    }

    /**
     * Returns true if this cell has crumbs belonging to any robot other than
     * <code>exceptForMe</code>.
     * 
     * @param exceptForMe The robot to exclude from the search.
     * @return Whether there are crumbs belonging to any other robots.
     */
    public boolean isCrumbyExceptFor(Robot exceptForMe) {
        return _trails.size() > 1
                || (_trails.size() == 1 && !_trails.contains(exceptForMe));
    }

    /**
     * Tests if another object is equal to this cell. True iff the other object
     * is an instance of StreetCell, and if the other cell has the same robots'
     * trails, and the same robot occupying it.
     * 
     * @param object The other object.
     * @return Whether the other object is equal to this one.
     */
    @Override
    public boolean equals(Object object) {
        if (!(object instanceof StreetCell)) {
            return false;
        }
        StreetCell cell = (StreetCell) object;
        return _trails.equals(cell._trails)
                && areEqual(_robotOccupying, cell._robotOccupying);
    }

    private boolean areEqual(Object o1, Object o2) {
        return (o1 == null ? o2 == null : o1.equals(o2));
    }

    /**
     * Returns this object's hash code.
     * 
     * @return This cell's hashCode.
     */
    @Override
    public int hashCode() {
        int result = 17;
        result = 37 * result + _trails.hashCode();
        result = 37 * result
                + (_robotOccupying == null ? 0 : _robotOccupying.hashCode());
        return result;
    }

    /**
     * Returns a String representation of this cell.
     * 
     * @return A String representation of this cell.
     */
    @Override
    public String toString() {
        StringBuilder result = new StringBuilder();
        result.append("StreetCell:");
        result.append(" trails: " + trailIDs());
        result.append(" occupier: "
                + (_robotOccupying == null ? "none" : ""
                        + _robotOccupying.getID()));
        result.append(" checked: " + _checked);

        return result.toString();
    }

    private SortedSet<Integer> trailIDs() {
        SortedSet<Integer> result = new TreeSet<Integer>();
        for (Robot robot : _trails) {
            result.add(robot.getID());
        }
        return result;
    }

    /**
     * Returns true iff this cell is checked.
     * 
     * @return Whether this cell is checked.
     */
    public boolean isChecked() {
        return _checked;
    }

    /**
     * Sets the checked value of this field.
     * 
     * @param checked The value to set.
     */
    public void setChecked(boolean checked) {
        _checked = checked;
    }

    /**
     * Returns true if this cell is an obstacle.  Returns true if and only if
     * this street is occupied.
     * 
     * @return Whether this cell is an obstacle.
     */
    public boolean isObstacle() {
        return isOccupied();
    }

    /**
     * Returns the message to be logged when a robot attempts to pass through
     * this when it's an obstacle.
     * 
     * @return The obstacle message.
     */
    public String getObstacleMessage() {
        return "Game Logic: Robot blocked by another robot.  "
                + "\"You shall not pass!\" it exclaims.";
    }

    /**
     * Returns the type of this cell.  Returns CellType.STREET.
     * 
     * @return The type of this cell.
     */
    public Sensor.CellType cellType() {
        return Sensor.CellType.STREET;
    }
}
