package backend;

/**
 * Singleton cell indicating a cell outside the playing field bounds (a wall
 * cell).  To get the instance of this class, call getInstance().
 * 
 * @author Matthew Justin
 */
public class WallCell implements Cell {

    /**
     * The single instance of this class.
     */
    private static final WallCell INSTANCE = new WallCell();

    /**
     * Private constructor so this class can't be instantiated.
     */
    private WallCell() {
        // can't be instantiated.
    }

    /**
     * Returns the singleton instance of this class.
     * 
     * @return The singleton instance.
     */
    public static WallCell getInstance() {
        return INSTANCE;
    }

    /**
     * Returns true if this cell is checked.  Always returns true.
     * 
     * @return Whether this cell is checked.
     */
    public boolean isChecked() {
        return true;
    }

    /**
     * Sets the checked value of this field. Always throws an
     * UnsupportedOperationException.
     * 
     * @param checked The value to set.
     * @throws UnsupportedOperationException Always thrown.
     */
    public void setChecked(boolean b) {
        throw new UnsupportedOperationException();
    }

    /**
     * Returns true if this cell is an obstacle.  Always returns true.
     * 
     * @return Whether this cell is an obstacle
     */
    public boolean isObstacle() {
        return true;
    }

    /**
     * Returns the message to be logged when a robot attempts to pass through
     * this when it's an obstacle.
     * 
     * @return The obstacle message.
     */
    public String getObstacleMessage() {
        return "Game Logic: Robot dines at the end of the universe.";
    }

    /**
     * Returns the type of this cell.  Returns CellType.WALL.
     * 
     * @return The type of this cell.
     */
    public Sensor.CellType cellType() {
        return Sensor.CellType.WALL;
    }

    /**
     * Returns a String representation of this cell.
     * 
     * @return A String representation of this cell.
     */
    public String toString() {
        return "Wall Cell";
    }
}
