package gui;

import java.awt.Color;
import java.awt.Graphics;

import javax.swing.JPanel;

import backend.GameListener;
import backend.GameLogic;

/**
 * JPanel which details turn information for a given game.  Will display the
 * current turn, the max number of turns, and whether the game is over.
 * 
 * @author Seth Sweep
 */
public class InfoPanel extends JPanel implements GameListener {
    /**
     * The height of the panel, in pixels.
     */
    public static final int PANEL_HEIGHT = 15;

    private static final Color BACKGROUND_COLOR = new Color(255, 255, 204);

    private GameLogic _gameLogic;

    /**
     * Creates a new InfoPanel for the given GameLogic. Upon construction this
     * will add itself as a listener to the game.
     * 
     * @param gameLogic The GameLogic serving as the model.
     */
    public InfoPanel(GameLogic gameLogic) {
        setBackground(BACKGROUND_COLOR);

        _gameLogic = gameLogic;
        _gameLogic.addListener(this);
    }

    public void paintComponent(Graphics graphics) {
        if (graphics == null) {
            return;
        }
        super.paintComponent(graphics);
        Color textColor;
        String message;
        if (_gameLogic.isGameOver()) {
            textColor = Color.red;
            message = gameOverMessage();
        } else {
            textColor = Color.black;
            message = gameInProgressMessage();
        }
        graphics.setColor(textColor);
        graphics.drawString(message, 5, 12);
        graphics.dispose();
    }

    private String gameInProgressMessage() {
        String result = String.format("Game in progress. Turn (%d/%d)",
                _gameLogic.getCurrentTurn() - 1, _gameLogic.getMaxTurns());
        return result;
    }

    private String gameOverMessage() {
        String result = "Game Over.";
        if (_gameLogic.getCurrentTurn() != _gameLogic.getMaxTurns()) {
            result += "  Game finished in " + _gameLogic.getCurrentTurn()
                    + " turns!";
        }
        return result;
    }

    /**
     * Called when the game state has changed. When this method is called, an
     * unspecified portion of the field has changed. Other details of the game
     * state may have changed as well (score, current move, game over).
     * 
     * Updates the display to reflect the current state of the game.
     */
    public void fireGameViewChanged() {
        paintComponent(getGraphics());
    }

    /**
     * Called when the game state has changed. This method is called to indicate
     * that only a given portion of the field has changed. Other details of the
     * game state may have changed as well (score, current move, game over).
     * 
     * Updates the display to reflect the current state of the game.
     * 
     * @param x The leftmost x coordinate of the changed area.
     * @param y The topmost y coordinate of the changed area.
     * @param width The width of the changed area.
     * @param height The height of the changed area.
     */
    public void fireGameViewChanged(int x, int y, int width, int height) {
        paintComponent(getGraphics());
    }
}
