package gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Polygon;
import java.awt.Shape;

import javax.swing.JPanel;

import backend.BuildingCell;
import backend.Cell;
import backend.Direction;
import backend.GameListener;
import backend.GameLogic;
import backend.PlayingField;
import backend.Robot;
import backend.StreetCell;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * JPanel which displays the playing field as a color-coded grid.
 * 
 * @author Seth Sweep
 */
public class PlayingFieldPanel extends JPanel implements GameListener {

    private PlayingField _playingField;

    private final Color UNOCCUPIED_BUILDING_COLOR = Color.green;

    private final Color ROAD_COLOR = Color.lightGray;

    private final Color CRUMBS_COLOR = Color.gray;

    private final static Color[] ROBOT_COLORS = { Color.blue, Color.red,
            Color.magenta, Color.orange };

    private final static Color DEFAULT_ROBOT_COLOR = Color.WHITE;

    /**
     * Creates a new PlayingFieldPanel for the given GameLogic. Upon
     * construction this will add itself as a listener to the game.
     * 
     * @param gameLogic The GameLogic serving as the model.
     */
    public PlayingFieldPanel(GameLogic gameLogic) {
        setBackground(ROAD_COLOR);
        gameLogic.addListener(this);
        _playingField = gameLogic.getPlayingField();
    }

    /**
     * Paints the component.
     * 
     * @param graphics The graphics being painted.
     */
    public void paintComponent(Graphics graphics) {
        // super.paintComponent(g);
        paintComponent((Graphics2D) graphics, 0, 0, _playingField.getWidth(),
                _playingField.getHeight());
    }

    /**
     * Paints a portion of the field. Coordinates measure in cells.
     * 
     * @param g The graphics object on which to paint.
     * @param startX The leftmost x coordinate.
     * @param startY The topmost y coordinate.
     * @param endX The width of the area to paint.
     * @param endY The height of the area to paint.
     */
    private void paintComponent(Graphics2D graphics, int startX, int startY,
            int endX, int endY) {
        if (graphics == null)
            return;
        // super.paintComponent(g);
        startX = Math.max(0, startX);
        endX = Math.min(endX, _playingField.getWidth() - 1);
        startY = Math.max(0, startY);
        endY = Math.min(endY, _playingField.getHeight() - 1);

        paintCells(graphics, startX, startY, endX, endY);
        graphics.dispose();
    }

    private void paintCells(Graphics2D graphics, int startX, int startY,
            int endX, int endY) {
        int fieldWidth = _playingField.getWidth();
        int fieldHeight = _playingField.getHeight();

        int panelWidth = getWidth();
        int panelHeight = getHeight();

        int shapeWidth = shapeSize(fieldWidth, panelWidth);
        int shapeHeight = shapeSize(fieldHeight, panelHeight);
        for (int x = startX; x <= endX; x++) {
            for (int y = startY; y <= endY; y++) {
                int shapeXpos = panelPosition(fieldWidth, panelWidth, x);
                int shapeYpos = panelPosition(fieldHeight, panelHeight, y);

                Cell cell = _playingField.getCell(x, y);
                renderCell(graphics, cell, shapeXpos, shapeYpos, shapeWidth,
                        shapeHeight);
            }
        }
    }

    private int shapeSize(double fieldSize, double panelSize) {
        return (int) (panelSize / fieldSize) + 1;
    }

    private int panelPosition(double fieldSize, double panelSize,
            double fieldPosition) {
        return (int) ((fieldPosition / fieldSize) * panelSize);
    }

    private void renderCell(Graphics2D graphics, Cell cell, int shapeXpos,
            int shapeYpos, int shapeWidth, int shapeHeight) {
        if (cell instanceof BuildingCell) {
            renderBuildingCell(graphics, (BuildingCell) cell, shapeXpos,
                    shapeYpos, shapeWidth, shapeHeight);
        } else if (cell instanceof StreetCell) {
            renderStreetCell(graphics, (StreetCell) cell, shapeXpos, shapeYpos,
                    shapeWidth, shapeHeight);
        }
    }

    private void renderBuildingCell(Graphics2D graphics, BuildingCell cell,
            int shapeXpos, int shapeYpos, int shapeWidth, int shapeHeight) {
        Color cellColor = this.UNOCCUPIED_BUILDING_COLOR;
        if (cell.isOwned()) {
            cellColor = robotColor(cell.getOwner());
        }
        cellColor = cellColor.brighter();
        for (int i = 1; i < cell.getPoints(); i++) {
            cellColor = cellColor.darker();
        }
        renderRectangularCell(graphics, shapeXpos, shapeYpos, shapeWidth,
                shapeHeight, cellColor);
    }

    private void renderStreetCell(Graphics2D graphics, StreetCell cell,
            int shapeXpos, int shapeYpos, int shapeWidth, int shapeHeight) {
        Color cellColor = cell.isCrumby() ? CRUMBS_COLOR : ROAD_COLOR;
        renderRectangularCell(graphics, shapeXpos, shapeYpos, shapeWidth,
                shapeHeight, cellColor);

        if (cell.isOccupied()) {
            renderRobot(graphics, cell.getRobotOccupying(), shapeXpos,
                    shapeYpos, shapeWidth, shapeHeight);
        }
    }

    private void renderRectangularCell(Graphics2D graphics, int shapeXpos,
            int shapeYpos, int shapeWidth, int shapeHeight, Color bCellColor) {
        graphics.setColor(bCellColor);
        graphics.fillRect(shapeXpos, shapeYpos, shapeWidth, shapeHeight);
    }

    private void renderRobot(Graphics2D graphics, Robot robot, int shapeXpos,
            int shapeYpos, int shapeWidth, int shapeHeight) {
        // May want to consider switching to a filled triangle outlined with
        // black via drawShape().
        Direction dir = robot.getDirection();
        Polygon outerWedge = createOuterWedge(dir, shapeXpos, shapeYpos,
                shapeWidth, shapeHeight);
        Polygon innerWedge = createInnerWedge(outerWedge, dir, shapeWidth,
                shapeHeight);

        drawFilledShape(graphics, Color.black, outerWedge);
        drawFilledShape(graphics, robotColor(robot), innerWedge);
    }

    private void drawFilledShape(Graphics2D graphics, Color color, Shape shape) {
        graphics.setColor(color);
        graphics.fill(shape);
    }

    private Polygon createOuterWedge(Direction robotDirection, int shapeXpos,
            int shapeYpos, int shapeWidth, int shapeHeight) {
        Point pointA = new Point();
        Point pointB = new Point();
        Point pointC = new Point();

        if (robotDirection == Direction.EAST) {
            pointA.x = 0;
            pointA.y = 0;
            pointB.x = 0;
            pointB.y = shapeHeight;
            pointC.x = shapeWidth;
            pointC.y = (shapeHeight / 2);
        } else if (robotDirection == Direction.NORTH) {
            pointA.x = 0;
            pointA.y = shapeHeight;
            pointB.x = shapeWidth;
            pointB.y = shapeHeight;
            pointC.x = (shapeWidth / 2);
            pointC.y = 0;
        } else if (robotDirection == Direction.WEST) {
            pointA.x = shapeWidth;
            pointA.y = shapeHeight;
            pointB.x = shapeWidth;
            pointB.y = 0;
            pointC.x = 0;
            pointC.y = (shapeHeight / 2);
        } else if (robotDirection == Direction.SOUTH) {
            pointA.x = shapeWidth;
            pointA.y = 0;
            pointB.x = 0;
            pointB.y = 0;
            pointC.x = (shapeWidth / 2);
            pointC.y = shapeHeight;
        }

        pointA.translate(shapeXpos, shapeYpos);
        pointB.translate(shapeXpos, shapeYpos);
        pointC.translate(shapeXpos, shapeYpos);

        return createPolygon(pointA, pointB, pointC);
    }

    private Polygon createInnerWedge(Polygon outerWedge,
            Direction robotDirection, int shapeWidth, int shapeHeight) {
        Point pointA = new Point(outerWedge.xpoints[0], outerWedge.ypoints[0]);
        Point pointB = new Point(outerWedge.xpoints[1], outerWedge.ypoints[1]);
        Point pointC = new Point(outerWedge.xpoints[2], outerWedge.ypoints[2]);

        int widthBorder = shapeWidth / 4;
        int heightBorder = shapeHeight / 4;

        if (robotDirection == Direction.EAST) {
            pointA.x += widthBorder;
            pointA.y += heightBorder;
            pointB.x += widthBorder;
            pointB.y -= heightBorder;
            pointC.x -= widthBorder;
            pointC.y = pointC.y;
        } else if (robotDirection == Direction.NORTH) {
            pointA.x += widthBorder;
            pointA.y -= heightBorder;
            pointB.x -= widthBorder;
            pointB.y -= heightBorder;
            pointC.x = pointC.x;
            pointC.y += heightBorder;
        } else if (robotDirection == Direction.WEST) {
            pointA.x -= widthBorder;
            pointA.y -= heightBorder;
            pointB.x -= widthBorder;
            pointB.y += heightBorder;
            pointC.x += widthBorder;
            pointC.y = pointC.y;
        } else if (robotDirection == Direction.SOUTH) {
            pointA.x -= widthBorder;
            pointA.y += heightBorder;
            pointB.x += widthBorder;
            pointB.y += heightBorder;
            pointC.x = pointC.x;
            pointC.y -= heightBorder;
        }
        return createPolygon(pointA, pointB, pointC);
    }

    private Polygon createPolygon(Point... points) {
        Polygon result = new Polygon();
        for (Point p : points) {
            result.addPoint(p.x, p.y);
        }
        return result;
    }

    private Color robotColor(Robot robot) {
        int player = robot.getID() - 1;
        if (player >= 0 && player < ROBOT_COLORS.length) {
            return ROBOT_COLORS[player];
        } else {
            return DEFAULT_ROBOT_COLOR;
        }
    }

    /**
     * Called when the game state has changed. When this method is called, an
     * unspecified portion of the field has changed. Other details of the game
     * state may have changed as well (score, current move, game over).
     * 
     * Updates the display to reflect the current state of the game.
     */
    public void fireGameViewChanged() {
        paintComponent(getGraphics());
    }

    /**
     * Called when the game state has changed. This method is called to indicate
     * that only a given portion of the field has changed. Other details of the
     * game state may have changed as well (score, current move, game over).
     * 
     * Updates the display of the indicated area to reflect the current state of
     * the game.
     * 
     * @param x The leftmost x coordinate of the changed area.
     * @param y The topmost y coordinate of the changed area.
     * @param width The width of the changed area.
     * @param height The height of the changed area.
     */
    public void fireGameViewChanged(int x, int y, int width, int height) {
        paintComponent((Graphics2D) getGraphics(), x, y, width, height);
    }
}
