import java.util.Arrays;
import java.util.List;

import junit.framework.TestCase;
import robots.ConstructorExceptionBot;
import robots.FigureEightAlgorithm;
import robots.PredictableTestRobot;
import algorithms.ManualControl;
import backend.RobotAlgorithmIF;

public class GameOptionsTest extends TestCase {
    private static final String THROW_EXCEPTIONS = "exceptions";

    public void testDefaultConstructor() {
        final String message = "The argument should have been null";
        GameOptions options = new GameOptions();
        assertEquals(message, null, options.algorithms);
        assertEquals(message, null, options.level);
        assertEquals(message, null, options.visible);
        assertEquals(message, null, options.delay);
        assertEquals(message, null, options.exceptions);
        assertEquals(message, null, options.verbose);
    }

    public void testSixArgumentConstructor() {
        RobotAlgorithmIF[] algorithms = new RobotAlgorithmIF[] {
                new PredictableTestRobot(), new ManualControl() };
        Integer level = 5;
        Boolean visible = false;
        Integer delay = 3;
        Boolean throwBotExceptions = true;
        Boolean verbose = null;
        Long seed = null;
        Boolean output = false;

        final String message = "The wrong argument was returned";
        GameOptions options = new GameOptions(algorithms, level, visible,
                delay, throwBotExceptions, verbose, seed, output);
        assertEquals(message, algorithms, options.algorithms);
        assertEquals(message, level, options.level);
        assertEquals(message, visible, options.visible);
        assertEquals(message, delay, options.delay);
        assertEquals(message, throwBotExceptions, options.exceptions);
        assertEquals(message, verbose, options.verbose);
    }

    public void testConstants() {
        List<String> expected = Arrays.asList("algorithms", "level", "visible",
                "delay", THROW_EXCEPTIONS, "verbose", "seed", "output");
        assertEquals("The wrong arguments were returned", expected,
                GameOptions.LEGAL_ARGUMENTS);
        try {
            GameOptions.LEGAL_ARGUMENTS.clear();
            fail("The arguments shouldn't be modifiable");
        } catch (UnsupportedOperationException e) {
            // this should happen
        }
    }

    public void testSetOptionToValue() {
        GameOptions options = new GameOptions();
        assertGetAndSetOptionValue(options, "algorithms",
                new RobotAlgorithmIF[] { new PredictableTestRobot() });
        assertGetAndSetOptionValue(options, "algorithms",
                new RobotAlgorithmIF[] { new FigureEightAlgorithm() });
        assertGetAndSetOptionValue(options, "level", 2);
        assertGetAndSetOptionValue(options, "visible", true);
        assertGetAndSetOptionValue(options, "delay", 30);
        assertGetAndSetOptionValue(options, THROW_EXCEPTIONS, null);
        assertGetAndSetOptionValue(options, "verbose", false);
        assertGetAndSetOptionValue(options, "seed", null);
        assertGetAndSetOptionValue(options, "seed", 10025555555L);
    }

    public void testGetAndSetOption() {
        GameOptions options = new GameOptions();
        assertGetAndSetOption(options, "algorithms",
                new RobotAlgorithmIF[] { new ManualControl() },
                "algorithms.ManualControl");
        assertGetAndSetOption(options, "algorithms", new RobotAlgorithmIF[] {
                new ManualControl(), new ManualControl() },
                "algorithms.ManualControl", "algorithms.ManualControl");
        assertGetAndSetOption(options, "level", 1, "1");
        assertGetAndSetOption(options, "level", -1, "-1");
        assertGetAndSetOption(options, "visible", false, "false");
        assertGetAndSetOption(options, "visible", false, "FalSE");
        assertGetAndSetOption(options, "visible", true, "true");
        assertGetAndSetOption(options, "visible", true, "tRUe");
        assertGetAndSetOption(options, "visible", true);
        assertGetAndSetOption(options, "delay", 4, "4");
        assertGetAndSetOption(options, THROW_EXCEPTIONS, false, "false");
        assertGetAndSetOption(options, THROW_EXCEPTIONS, true, "true");
        assertGetAndSetOption(options, THROW_EXCEPTIONS, true);
        assertGetAndSetOption(options, "verbose", false, "false");
        assertGetAndSetOption(options, "verbose", true, "true");
        assertGetAndSetOption(options, "verbose", true);
        assertGetAndSetOption(options, "seed", 1002555555599999L,
                "1002555555599999");
    }

    public void testIllegalSetOptionToValue() {
        GameOptions options = new GameOptions();
        assertIllegalSetOptionToValue(options, "algorithms",
                new PredictableTestRobot());
        assertIllegalSetOptionToValue(options, "level", "1");
        assertIllegalSetOptionToValue(options, "level", "1002555555599999");
        assertIllegalSetOptionToValue(options, "level", false);
        assertIllegalSetOptionToValue(options, "visible", "false");
        assertIllegalSetOptionToValue(options, "visible", 1);
        assertIllegalSetOptionToValue(options, "delay", "1");
        assertIllegalSetOptionToValue(options, "delay", "1002555555599999");
        assertIllegalSetOptionToValue(options, THROW_EXCEPTIONS, "false");
        assertIllegalSetOptionToValue(options, "verbose", 1);
        assertIllegalSetOptionToValue(options, "LEGAL_ARGUMENTS",
                "robots");
        assertIllegalSetOptionToValue(options, "bananaphone", null);
    }

    public void testAlgorithmWithConstructorException() {
        try {
            new GameOptions().setOptionToValue("algorithm",
                    new ConstructorExceptionBot());
            fail("A runtime exception should have been thrown.");
        } catch (RuntimeException e) {
            // this should happen.
        }

        try {
            new GameOptions().setOption("algorithms",
                    "robots.ConstructorExceptionBot");
            fail("A runtime exception should have been thrown.");
        } catch (RuntimeException e) {
            assertEquals("The wrong exception was thrown",
                    "This robot throws exceptions when constructed.", e
                            .getCause().getMessage());
        }
    }

    public void testIllegalGetOption() {
        GameOptions options = new GameOptions();
        assertIllegalGetOption(options, "LEGAL_ARGUMENTS");
        assertIllegalGetOption(options, "bananaphone");
    }

    public void testIllegalSetOption() {
        GameOptions options = new GameOptions();
        assertIllegalSetOption(options, "algorithms");
        assertIllegalSetOption(options, "algorithms", "backend.notarobot");
        assertIllegalSetOption(options, "algorithms", "");
        assertIllegalSetOption(options, "algorithms", "null");
        assertIllegalSetOption(options, "algorithms", "backend.Move");
        assertIllegalSetOption(options, "algorithms",
                "backend.RobotAlgorithmIF");
        assertIllegalSetOption(options, "algorithms", "backend.RobotAlgorithm");
        assertIllegalSetOption(options, "level", "false");
        assertIllegalSetOption(options, "level", "one");
        assertIllegalSetOption(options, "level", "null");
        assertIllegalSetOption(options, "level", "1", "18");
        assertIllegalSetOption(options, "visible", "0");
        assertIllegalSetOption(options, "visible", "");
        assertIllegalSetOption(options, "visible", "null");
        assertIllegalSetOption(options, "visible", "false", "true");
        assertIllegalSetOption(options, "delay", "false");
        assertIllegalSetOption(options, THROW_EXCEPTIONS, "1");
        assertIllegalSetOption(options, "verbose", "1");
        assertIllegalSetOption(options, "LEGAL_ARGUMENTS", "robots");
        assertIllegalSetOption(options, "bananaphone", "null");
    }

    private void assertIllegalSetOption(GameOptions options, String option,
            String... value) {
        try {
            options.setOption(option, value);
            fail("An exception should have been thrown.");
        } catch (IllegalArgumentException e) {
            // this should happen
        }
    }

    private void assertIllegalSetOptionToValue(GameOptions options,
            String option, Object value) {
        try {
            options.setOptionToValue(option, value);
            fail("An exception should have been thrown.");
        } catch (IllegalArgumentException e) {
            // this should happen
        }
    }

    private void assertIllegalGetOption(GameOptions options, String option) {
        try {
            options.getOption(option);
            fail("An exception should have been thrown.");
        } catch (IllegalArgumentException e) {
            // this should happen
        }
    }

    public void testToString() {
        RobotAlgorithmIF[] algorithms = new RobotAlgorithmIF[] {
                new PredictableTestRobot(), new ManualControl() };
        Integer level = 5;
        Boolean visible = false;
        Integer delay = 3;
        Boolean throwBotExceptions = true;
        Boolean verbose = null;
        Long seed = null;
        Boolean output = false;

        final String message = "The wrong String was returned";
        GameOptions options = new GameOptions(algorithms, level, visible,
                delay, throwBotExceptions, verbose, seed, output);
        String expected = "GameOptions: algorithms: "
                + "[PredictableTestRobot, Manual Control]; level: 5; "
                + "visible: false; delay: 3; exceptions: true; "
                + "verbose: null; seed: null; output: false";
        assertEquals(message, expected, options.toString());
    }

    private void assertGetAndSetOption(GameOptions options, String option,
            Object expected, String... values) {
        options.setOption(option, values);
        assertGetOption(options, option, expected);
        options.setOptionToValue(option, null);
        assertGetAndSetOption(options, option, expected, Arrays.asList(values));
    }

    private void assertGetAndSetOption(GameOptions options, String option,
            Object expected, List<String> values) {
        options.setOption(option, values);
        assertGetOption(options, option, expected);
    }

    private void assertGetOption(GameOptions options, String option,
            Object expected) {
        if (expected.getClass().equals(RobotAlgorithmIF[].class)) {
            Object actual = options.getOption(option);
            if (!(actual instanceof RobotAlgorithmIF[])) {
                fail("The returned value should have been a RobotAlgorithmIF[], but was a "
                        + actual.getClass());
            }
            RobotAlgorithmIF[] expectedArray = (RobotAlgorithmIF[]) expected;
            RobotAlgorithmIF[] actualArray = (RobotAlgorithmIF[]) actual;
            for (int i = 0; i < expectedArray.length; i++) {
                assertEquals("The value wasn't set properly", expectedArray[i]
                        .getClass(), actualArray[i].getClass());
            }
        } else {
            assertEquals("The value wasn't set properly", expected, options
                    .getOption(option));
        }
    }

    private void assertGetAndSetOptionValue(GameOptions options, String option,
            Object value) {
        options.setOptionToValue(option, value);
        assertEquals("The value wasn't set properly", value, options
                .getOption(option));
    }
}
