package algorithms;

import java.awt.Point;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.io.Reader;
import java.io.StringBufferInputStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.io.Writer;

import junit.framework.AssertionFailedError;
import junit.framework.TestCase;
import backend.Move;
import backend.PlayingField;
import backend.Robot;
import backend.RobotAlgorithmIF;
import backend.SensorGenerator;
import backend.SensorGroup;

public class ManualControlTest extends TestCase {
    @SuppressWarnings("deprecation")
    public void testDefaultConstructor() {
        /*
         * Redirecting input streams so we can make sure that the default
         * constructor is reading from <code>System.in</code> and writing to
         * <code>System.out</code>, as I see no obvious better way of doing
         * this.
         */
        InputStream previousIn = System.in;
        OutputStream previousOut = System.out;
        InputStream in = new StringBufferInputStream("6\n");
        OutputStream out = new ByteArrayOutputStream();
        System.setIn(in);
        System.setOut(new PrintStream(out));
        try {
            testOneMove(out);
        } catch (AssertionFailedError e) {
            throw e;
        } catch (RuntimeException e) {
            throw e;
        } finally {
            // Set the streams back, so that they are set properly for
            // successive tests.
            System.setIn(previousIn);
            System.setOut(new PrintStream(previousOut));
        }
    }

    private void testOneMove(OutputStream out) {
        ManualControl algorithm = new ManualControl();

        SensorGroup group = new SensorGroup(new PlayingField(1, 1),
                new Robot(0));

        assertEquals("The wrong move was returned", Move.TURN_RIGHT,
                algorithm.getMove(group));
        assertEquals("The wrong string was returned",
                expectedOutput(group), out.toString());
    }
    
    public void testGetMove() {
        String[] inputStrings = {
                "5",
                "8",
                "4",
                "6",
                "856",
                "6ll",
                "b",
                "ll",
                "",
        };
        Move[] expectedMoves = {
                Move.AHEAD_WITH_CRUMBS,
                Move.AHEAD_WITHOUT_CRUMBS,
                Move.TURN_LEFT,
                Move.TURN_RIGHT,
                Move.AHEAD_WITHOUT_CRUMBS,
                Move.TURN_RIGHT,
                Move.SKIP,
                Move.SKIP,
                Move.SKIP,
        };
        
        Reader reader = new StringReader(toInputString(inputStrings));
        StringWriter writer = new StringWriter();

        RobotAlgorithmIF algorithm = new ManualControl(reader, writer);

        testCorrectMoves(expectedMoves, writer, algorithm);
    }

    private void testCorrectMoves(Move[] expectedMoves, StringWriter writer, RobotAlgorithmIF algorithm) {
        PlayingField field = new PlayingField(1, 1);
        Robot robot = new Robot(0);
        SensorGenerator generator = new SensorGenerator(field, robot);
        for (Move expectedMove : expectedMoves) {
            Point location = robot.getLocation();
            location.translate(2, 2);
            field.setRobotLocation(robot, location);
            SensorGroup sensors = generator.getSensors();
            Move actualMove = algorithm.getMove(sensors);
            assertEquals("The wrong move was returned", expectedMove,
                    actualMove);
            StringBuffer sensorData = writer.getBuffer();
            assertEquals("The wrong sensor data was displayed",
                    expectedOutput(sensors), sensorData.toString());
            sensorData.setLength(0);
        }
    }

    public void testGetName() {
        ManualControl algorithm = new ManualControl();
        assertEquals("The wrong string was returned", "Manual Control",
                algorithm.toString());
    }

    public void testToString() {
        ManualControl algorithm = new ManualControl();
        assertEquals("The wrong string was returned", algorithm.getName(),
                algorithm.toString());
    }
    
    public void testIOExceptionHandling() {
        Reader reader = new ExceptionReader();
        Writer writer = new StringWriter();
        ManualControl algorithm = new ManualControl(reader, writer);
        
        SensorGroup group = new SensorGroup(new PlayingField(1, 1),
                new Robot(0));
        
        try {
            algorithm.getMove(group);
            fail("A RuntimeException should have been thrown");
        } catch (RuntimeException e) {
            assertEquals("The thrown Exception had the wrong cause",
                    IOException.class, e.getCause().getClass());
        }
        
        reader = new StringReader("3\n");
        writer = new ExceptionWriter();
        algorithm = new ManualControl(reader, writer);

        try {
            algorithm.getMove(group);
            fail("A RuntimeException should have been thrown");
        } catch (RuntimeException e) {
            assertEquals("The thrown Exception had the wrong cause",
                    IOException.class, e.getCause().getClass());
        }
    }

    private String expectedOutput(SensorGroup sensors) {
        StringBuilder result = new StringBuilder();
        result.append("__________SENSORS__________\n");
        result.append(sensors + "\n");
        result.append("AWAITING MOVE...\n");
        return result.toString();
    }

    private String toInputString(String[] inputStrings) {
        StringBuilder result = new StringBuilder();
        for (String string : inputStrings) {
            result.append(string);
            result.append("\n");
        }
        return result.toString();
    }

    private static class ExceptionWriter extends Writer {
        @Override
        public void write(char[] cbuf, int off, int len) throws IOException {
            throw new IOException("This writer throws exceptions");
        }

        @Override
        public void flush() throws IOException {
        }

        @Override
        public void close() throws IOException {
        }
    }

    private static class ExceptionReader extends Reader {
        @Override
        public int read(char[] cbuf, int off, int len) throws IOException {
            throw new IOException("This reader throws exceptions");
        }

        @Override
        public void close() throws IOException {
        }
    }
}