package backend;

import java.util.List;
import java.util.Random;

import junit.framework.TestCase;
import robots.ExceptionBot;
import robots.FigureEightAlgorithm;
import robots.InnerEnclosureAlgorithm;
import robots.IrregularEnclosureTestAlgorithm;
import robots.IrregularEnclosureTestAlgorithm2;
import robots.IrregularTestAlg2;
import robots.LargeEnclosureTestAlgorithm;
import robots.MultipleRobotsTestAlgorithm1;
import robots.MultipleRobotsTestAlgorithm2;
import robots.NullMoveRobot;
import robots.PredictableTestRobot;
import testutils.TestUtils;
import exceptions.NotEnoughPlayersException;
import exceptions.TooManyPlayersException;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * 
 * 
 * @author Seth Sweep
 */
public class GameLogicTest extends TestCase {
	
	public void testConstructor() {
		RobotAlgorithmIF[]  algorithms = {new PredictableTestRobot(), new PredictableTestRobot()};
		for (int level = -1; level < 5; level++) {
			GameLogic logic = new GameLogic(algorithms, level);
			assertEquals("The wrong level was returned", level, logic.getPlayingField().getLevel());
            assertEquals("The wrong level was returned", level, logic.getLevel());
			assertEquals("The wrong max moves was returned", 5000, logic.getMaxTurns());
			assertEquals("The wrong current move was returned", 1, logic.getCurrentTurn());
			assertFalse("The game should not have been over", logic.isGameOver());
            TestUtils.assertArraysEqual("The initial scores should be 0",
                    new int[] { 0, 0 }, logic.getScores());
		}
	}
	
	public void testConstructorWithInvalidArgumetns() {
		try {
			new GameLogic(null);
			fail("A NullPointerException should have been thrown");
		} catch (NullPointerException e) {
			//this should happen
		}
		try {
			new GameLogic(new RobotAlgorithmIF[]{});
			fail("A NotEnoughPlayersException should have been thrown");
		} catch (NotEnoughPlayersException e) {
			//this should happen
		}
		RobotAlgorithmIF[] algorithms = new RobotAlgorithmIF[4 * 51 - 3];
		for (int i = 0; i < algorithms.length; i++) {
			algorithms[i] = new PredictableTestRobot();
		}
		try {
			new GameLogic(algorithms);
			fail("A TooManyPlayersException should have been thrown");
		} catch (TooManyPlayersException e) {
			//this should happen
		}
		RobotAlgorithmIF[] algorithms2 = new RobotAlgorithmIF[4 * 51- 4];
		System.arraycopy(algorithms, 0, algorithms2, 0, algorithms2.length);
        new GameLogic(algorithms2);

        try {
            new GameLogic(new RobotAlgorithmIF[]{new PredictableTestRobot(), null});
            fail("A NullPointerException should have been thrown");
        } catch (NullPointerException e) {
            //this should happen
        }
	}

	public void testEnclosure() {
		RobotAlgorithmIF[] algs = { new LargeEnclosureTestAlgorithm() };
		GameLogic gl = new GameLogic(algs);
		PlayingField field = gl.getPlayingField();
        List<Robot> robots = field.getRobots();
		field.setRobot(robots.get(0), Direction.SOUTH, 0, 0);
		int[] endGameResult = gl.execute();
		assertEquals(
				"The robot should have enclosed a building and scored 2 points.",
				2, robots.get(0).getScore());
		TestUtils.assertArraysEqual("The wrong end-game result was returned",
				new int[] { 2 }, endGameResult);
	}

	public void testIrregularEnclosure()  {
		RobotAlgorithmIF[] algs = { new IrregularEnclosureTestAlgorithm(),
				new IrregularTestAlg2() };

		GameLogic gl = new GameLogic(algs);
		PlayingField field = gl.getPlayingField();
        List<Robot> robots = field.getRobots();
		field.setRobot(robots.get(0), Direction.NORTH, 4, 50);
		field.setRobot(robots.get(1), Direction.EAST, 0, 0);
		int[] endGameResult = gl.execute();
		assertEquals(
				"The robot should have enclosed buildings and scored 7 points.",
				7, robots.get(0).getScore());
		assertEquals(
				"The robot should have enclosed buildings and scored 15 points.",
				15, robots.get(1).getScore());
		TestUtils.assertArraysEqual("The wrong end-game result was returned",
				new int[] { 7, 15 }, endGameResult);

	}

    /**
     * Tests an enclosure that broke due to one of my (MJ's) optimizations.
     */
    public void testEnclosure2() {
        RobotAlgorithmIF[] algs = { new IrregularEnclosureTestAlgorithm2() };

        GameLogic gl = new GameLogic(algs);
        PlayingField field = gl.getPlayingField();
        List<Robot> robots = field.getRobots();
        int[] endGameResult = gl.execute();
        assertEquals(
                "The robot should have enclosed buildings and scored 2 points.",
                2, robots.get(0).getScore());
        TestUtils.assertArraysEqual("The wrong end-game result was returned",
                new int[] { 2 }, endGameResult);
    }

    public void testFigure8Enclosure() {
        RobotAlgorithmIF[] algs = { new FigureEightAlgorithm() };
        
        GameLogic gl = new GameLogic(algs);
        PlayingField field = gl.getPlayingField();
        List<Robot> robots = field.getRobots();
        int[] endGameResult = gl.execute();
        assertEquals(
                "The robot should have enclosed buildings and scored 3 points.",
                3, robots.get(0).getScore());
        TestUtils.assertArraysEqual("The wrong end-game result was returned",
                new int[] { 3 }, endGameResult);
    }
    
    public void testInnerEnclosure() {
        RobotAlgorithmIF[] algs = { new InnerEnclosureAlgorithm() };
        
        GameLogic gl = new GameLogic(algs);
        PlayingField field = gl.getPlayingField();
        List<Robot> robots = field.getRobots();
        int[] endGameResult = gl.execute();
        assertEquals(
                "The robot should have enclosed buildings and scored 18 points.",
                18, robots.get(0).getScore());
        TestUtils.assertArraysEqual("The wrong end-game result was returned",
                new int[] { 18 }, endGameResult);
    }
    
	public void testMultipleRobots() {
		RobotAlgorithmIF[] algs = { new MultipleRobotsTestAlgorithm1(),
				new MultipleRobotsTestAlgorithm2() };
		GameLogic gl = new GameLogic(algs);
		PlayingField field = gl.getPlayingField();
        List<Robot> robots = field.getRobots();
		field.setRobot(robots.get(0), Direction.NORTH, 30, 50);
		field.setRobot(robots.get(1), Direction.NORTH, 28, 50);
		int[] endGameResult = gl.execute();
		assertEquals(
				"The robot should have enclosed buildings and scored 3 points.",
				3, robots.get(0).getScore());
		assertEquals(
				"The robot should have not enclosed buildings and scored no points.",
				0, robots.get(1).getScore());
		TestUtils.assertArraysEqual("The wrong end-game result was returned",
				new int[] { 3, 0 }, endGameResult);
	}
	
	//TODO write a test to verify correct behavior when one robot bumps into another.

	public void testRandomness() throws InterruptedException {
		RobotAlgorithmIF[] algorithms = { new PredictableTestRobot(),
				new PredictableTestRobot(), new PredictableTestRobot(),
				new PredictableTestRobot(), new PredictableTestRobot(),
				new PredictableTestRobot(), };
		long l = 555846155;
		PlayingField f1 = new GameLogic(algorithms, 3, new Random(l))
				.getPlayingField();
		Thread.sleep(50);
		PlayingField f2 = new GameLogic(algorithms, 3, new Random(l))
				.getPlayingField();
		Thread.sleep(50);
		PlayingFieldTest.assertFieldsEqual(f1, f2, true);

		f2 = new GameLogic(algorithms, 3, new Random(l + 1)).getPlayingField();
		PlayingFieldTest.assertFieldsEqual(f1, f2, false);

		f1 = new GameLogic(algorithms, 3).getPlayingField();
		Thread.sleep(50);
		f2 = new GameLogic(algorithms, 3).getPlayingField();
		PlayingFieldTest.assertFieldsEqual(f1, f2, false);
	}
	
	public void testTurnByTurn() {
		PredictableTestRobot algorithm = new PredictableTestRobot();
		RobotAlgorithmIF[] algorithms = {algorithm};
		GameLogic logic = new GameLogic(algorithms);
		MockListener mockListener = new MockListener(algorithm, logic);
		logic.addListener(mockListener);
		logic.execute();
		assertEquals(5000, mockListener.getTurn());
		assertEquals(5000, algorithm.getMovesMade());
	}

    public void testMultiBotTurnByTurn() {
        final int expectedTurns = 4439;
        PredictableTestRobot algorithm = new PredictableTestRobot();
        RobotAlgorithmIF[] algorithms = {algorithm, new PredictableTestRobot()};
        GameLogic logic = new GameLogic(algorithms);
        MockListener mockListener = new MockListener(algorithm, logic);
        logic.addListener(mockListener);
        logic.execute();
        assertEquals(expectedTurns, mockListener.getTurn());
        assertEquals(expectedTurns, algorithm.getMovesMade());
    }

    /**
     * Verifies that a robot returning a null move doesn't crash the game.
     */
    public void testNullMove() {
        NullMoveRobot algorithm = new NullMoveRobot();
        PredictableTestRobot predictableRobot = new PredictableTestRobot();
        RobotAlgorithmIF[] algorithms = { algorithm, predictableRobot };
        GameLogic logic = new GameLogic(algorithms);
        int[] expected = new int[] { 0, 352 };
        TestUtils.assertArraysEqual("The wrong scores were returned", logic
                .execute(), expected);
    }

    /**
     * Verifies that a robot throwing an exception move doesn't crash the game.
     */
    public void testExceptionInAlgorithm() {
        ExceptionBot algorithm = new ExceptionBot();
        PredictableTestRobot predictableRobot = new PredictableTestRobot();
        RobotAlgorithmIF[] algorithms = { algorithm, predictableRobot };
        GameLogic logic = new GameLogic(algorithms);
        int[] expected = new int[] { 0, 352 };
        TestUtils.assertArraysEqual("The wrong scores were returned", logic
                .execute(), expected);
    }
    
    public void testExecuteOnlyCallableOnce() {
        PredictableTestRobot algorithm = new PredictableTestRobot();
        RobotAlgorithmIF[] algorithms = {algorithm};
        GameLogic logic = new GameLogic(algorithms);
        logic.execute();
        try {
            logic.execute();
            fail("An IllegalStateException should have been thrown");
        } catch (IllegalStateException e) {
            //This should happen
        }
    }
    
    //TODO Write test to verify correct robot position/direction after each move.
    //TODO Write test to verify correct getListener is called on a robot-by-robot basis.
    //TODO Write test for gameLogicReport()
    //TODO Write test to verify that setVerbose() calles Referee.setVerbose()
}
