package backend;

import java.awt.Point;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;

import junit.framework.TestCase;
import testutils.TestUtils;
import exceptions.MultipleRobotsInOnePlaceException;
import exceptions.NotEnoughPlayersException;
import exceptions.TooManyPlayersException;

/**
 * Algorithms and Compubility -- Project 2 (aka Kung Foo Robots)
 * 
 * This tests that the playing field is being built properly.
 * 
 * @author Seth Sweep
 */
public class PlayingFieldTest extends TestCase {

    PlayingField _playingField;

    /**
     * Constructor for PlayingFieldTest.
     * 
     * @param arg0
     */
    public PlayingFieldTest(String arg0) throws Exception {
        super(arg0);
        _playingField = new PlayingField(4, 4);
    }

    public void testPlayingFieldLevel1() {
        testStandardPlayingField(1);
    }

    public void testPlayingFieldLevel2() {
        PlayingField field = new PlayingField(4, 2);
        assertEquals("The wrong level was returned", 2, field.getLevel());
        assertCorrectDimensions(field, 51, 51);
        testCommonPlayingFieldGrid(field);
        testHigherLevelGrid(field, 50);
        testGridForRandomness(2, 50);
        assertAllCellsUnchecked(field);
        assertFalse("There should have been unclaimed buildings", field
                .allBuildingsClaimed());
    }

    public void testPlayingFieldLevel3() {
        PlayingField field = new PlayingField(4, 3);
        assertEquals("The wrong level was returned", 3, field.getLevel());
        assertCorrectDimensions(field, 61, 81);
        testCommonPlayingFieldGrid(field);
        testHigherLevelGrid(field, 200);
        testGridForRandomness(3, 200);
        assertAllCellsUnchecked(field);
        assertFalse("There should have been unclaimed buildings", field
                .allBuildingsClaimed());
    }

    public void testPlayingFieldDefault() {
        testStandardPlayingField(-1);
        testStandardPlayingField(4);
        testStandardPlayingField(20);
    }

    public void testNonrandomInitialRobotPositions() {
        for (int players = 1; players <= 4; players++) {
            for (int level = 1; level <= 4; level++) {
                PlayingField field = new PlayingField(players, level);
                assertInitialRobotPositions(field);
            }
        }
    }

    public void testRandomInitialRobotPositions() {
        Random random = new Random();
        List<Point> points = new ArrayList<Point>();
        List<Point> fullSet = new ArrayList<Point>();
        int players = 50;
        int numTests = 1000;
        for (int i = 1; i < 50; i++) {
            fullSet.add(new Point(0, i));
            fullSet.add(new Point(50, i));
            fullSet.add(new Point(i, 0));
            fullSet.add(new Point(i, 50));
        }

        for (int i = 0; i < numTests; i++) {
            PlayingField field = new PlayingField(players, 1, random);
            for (Robot robot : field.getRobots().subList(4,
                    field.getRobots().size())) {
                Point p = robot.getLocation();
                points.add(p);
                assertInitalDirectionCorrect(p, robot.getDirection());
            }
        }

        TestUtils.verifyRandomness(points, fullSet);
    }

    public void testMaximumNumberOfRobots() {
        int players = 2 * 51 + 2 * 51 - 4;
        PlayingField field = new PlayingField(players, 1);
        assertInitialRobotPositions(field);
        Set<Point> positions = new HashSet<Point>();
        for (Robot robot : field.getRobots()) {
            Point p = robot.getLocation();
            if (positions.contains(p)) {
                fail("Two robots were placed in the same square");
            }
            positions.add(p);
        }
    }

    public void testIllegalNumberOfPlayers() {
        try {
            new PlayingField(0, 1);
            fail("A NotEnoughPlayersException should have been thrown");
        } catch (NotEnoughPlayersException e) {
            // This should happen
        }

        try {
            new PlayingField(-1, 1);
            fail("A NotEnoughPlayersException should have been thrown");
        } catch (NotEnoughPlayersException e) {
            // This should happen
        }

        new PlayingField(2 * 51 + 2 * 51 - 4, 1);

        try {
            new PlayingField(2 * 51 + 2 * 51 - 3, 1);
            fail("A TooManyPlayersException should have been thrown");
        } catch (TooManyPlayersException e) {
            // This should happen
        }

        new PlayingField(2 * 61 + 2 * 81 - 4, 3);

        try {
            new PlayingField(2 * 61 + 2 * 81 - 3, 3);
            fail("A NotEnoughPlayersException should have been thrown");
        } catch (TooManyPlayersException e) {
            // This should happen
        }
    }

    public void testSetAndGetRobot() {
        Robot robot0 = new Robot(0);
        Robot robot1 = new Robot(1);
        Robot robot2 = new Robot(2);
        Robot robot3 = new Robot(3);
        testGetAndSetRobot(robot0, Direction.SOUTH, 1, 0);
        testGetAndSetRobot(robot1, Direction.WEST, 0, 1);
        testGetAndSetRobot(robot2, Direction.SOUTH, 2, 2);

        testGetAndSetRobot(robot1, Direction.WEST, 0, 1);
        testGetAndSetRobot(robot1, Direction.EAST, 0, 1);

        try {
            _playingField.setRobot(robot3, Direction.NORTH, 2, 2);
            fail("MultipleRobotsInOnePlaceException should have been thrown");
        } catch (MultipleRobotsInOnePlaceException e) {
            // this should happen!
        }

        try {
            _playingField.setRobot(robot3, Direction.NORTH, 1, 1);
            fail("An IllegalStateException should have been thrown when attempting to place a robot on a building.");
        } catch (IllegalStateException e) {
        }
    }

    /**
     * Tests that the random seed thing works for PlayingFields.
     * 
     * @throws TooManyPlayersException
     * @throws NotEnoughPlayersException
     * @throws InterruptedException
     */
    public void testPlayingFieldRandomness() throws InterruptedException {
        long l = 555846155;
        PlayingField f1 = new PlayingField(2, 2, new Random(l));
        Thread.sleep(50);
        PlayingField f2 = new PlayingField(2, 2, new Random(l));
        Thread.sleep(50);
        PlayingField f3 = new PlayingField(2, 2, new Random(l));
        assertFieldsEqual(f1, f2, true);
        assertFieldsEqual(f1, f3, true);

        f2 = new PlayingField(2, 2, new Random(l + 1));
        assertFieldsEqual(f1, f2, false);

        f1 = new PlayingField(2, 2);
        Thread.sleep(50);
        new PlayingField(2, 2);
        assertFieldsEqual(f1, f2, false);
    }

    //TODO Write test for Point version of getCell.
    public void testGetCell() {
        BuildingCell buildingCell = (BuildingCell) _playingField
                .getCell(1, 1);
        buildingCell.setOwner(new Robot(1));
        buildingCell.setPoints(20);
        buildingCell.setChecked(true);
        assertEquals("The cell should be the same", buildingCell, _playingField
                .getCell(1, 1));

        StreetCell streetCell = (StreetCell) _playingField
                .getCell(2, 2);
        streetCell.setRobotTrail(new Robot(1), true);
        assertEquals("The cell should be the same", streetCell, _playingField
                .getCell(2, 2));

        int[] values = { -1, 0, 1, 50, 51, 100 };

        for (int i : values) {
            for (int j : values) {
                boolean wallExpected = (i == -1 || j == -1 || i > 50 || j > 50);

                Cell cell = _playingField.getCell(i, j);

                assertEquals(
                        "The returned cell should(n't) have been a WallCell",
                        wallExpected, WallCell.getInstance() == cell);
            }
        }

        assertSame("The wall instance should have been returned", WallCell
                .getInstance(), _playingField.getCell(51, 51));
    }

    public void testClearChecked() {
        _playingField.getCell(3, 3).setChecked(true);
        _playingField.getCell(3, 9).setChecked(true);
        _playingField.getCell(20, 20).setChecked(true);

        _playingField.clearChecked();

        assertAllCellsUnchecked(_playingField);
    }

    public void testAllBuildingsClaimed() {
        for (int x = 1; x < _playingField.getWidth(); x += 2) {
            for (int y = 1; y < _playingField.getHeight(); y += 2) {
                BuildingCell cell = ((BuildingCell) _playingField.getCell(x, y));
                cell.setOwner(new Robot(2));
                boolean hasUnclaimedBuildings = !_playingField
                        .allBuildingsClaimed();
                boolean expected = !(y == _playingField.getWidth() - 2 && x == _playingField
                        .getHeight() - 2);
                assertEquals("The wrong value was returned", expected,
                        hasUnclaimedBuildings);
            }
        }
        BuildingCell cell = ((BuildingCell) _playingField.getCell(1, 1));
        cell.setOwner(null);
        assertFalse("The field should have had unclaimed buildings",
                _playingField.allBuildingsClaimed());
    }

    public void testNumPlayers() {
        for (int i = 1; i < 20; i++) {
            PlayingField field = new PlayingField(i, 1);
            assertEquals("The wrong number of players was returned", i, field
                    .getRobots().size());
        }
    }

    /**
     * Asserts that the initial direction of a randomly placed robot is correct.
     * 
     * @param x
     * @param y
     * @param robotDirection
     */
    private void assertInitalDirectionCorrect(Point location,
            Direction robotDirection) {
        int x = location.x;
        int y = location.y;
        Direction expected = null;
        if (y == 0) {
            expected = Direction.SOUTH;
        } else if (y == _playingField.getHeight() - 1) {
            expected = Direction.NORTH;
        } else if (x == 0) {
            expected = Direction.EAST;
        } else if (x == _playingField.getWidth() - 1) {
            expected = Direction.WEST;
        } else {
            fail("An invalid starting point was given");
        }
        assertEquals("The wrong starting direction was returned for robot at ("
                + x + ", " + y + ")", expected, robotDirection);
    }

    public static void assertAllCellsUnchecked(PlayingField field) {
        for (int x = 0; x < field.getWidth(); x++) {
            for (int y = 0; y < field.getHeight(); y++) {
                Cell cell = field.getCell(x, y);
                assertFalse("The cell should be unchecked", cell.isChecked());
            }
        }
    }

    /**
     * Tests that two PlayingFields are equal (or unequal).
     * 
     * @param f1 The first PlayingField
     * @param f2 The second PlayingField
     * @param equal True if you want to check for equality, false if you want to
     *        test for inequality.
     */
    public static void assertFieldsEqual(PlayingField f1, PlayingField f2,
            boolean equal) {
        final String msg = "The two fields %s and %s should %shave been equal";
        String not = (equal ? "" : "not ");
        String message = String.format(msg, f1, f2, not);
        if (equal) {
            assertTrue(message, fieldsEqual(f1, f2));
        } else {
            assertFalse(message, fieldsEqual(f1, f2));
        }
    }

    private static boolean fieldsEqual(PlayingField f1, PlayingField f2) {
        if (f1.getHeight() != f2.getHeight() || f1.getWidth() != f2.getWidth()) {
            return false;
        }
        for (int y = 0; y < f1.getHeight(); y++) {
            for (int x = 0; x < f1.getWidth(); x++) {
                if (f1.getCell(x, y) instanceof StreetCell
                        && f2.getCell(x, y) instanceof StreetCell) {
                    StreetCell cell1 = (StreetCell) f1.getCell(x, y);
                    StreetCell cell2 = (StreetCell) f2.getCell(x, y);
                    if (!sameRobotID(cell1.getRobotOccupying(), cell2
                            .getRobotOccupying())) {
                        return false;
                    }
                } else if (!(f1.getCell(x, y)).equals(f2.getCell(x, y))) {
                    return false;
                }
            }
        }
        return true;
    }

    private static boolean sameRobotID(Robot r1, Robot r2) {
        if (r1 == null & r2 == null) {
            return true;
        }
        if (r1 == null || r2 == null) {
            return false;
        }
        return r1.getID() == r2.getID();
    }

    private void testGetAndSetRobot(Robot robot, Direction dir, int x, int y) {
        _playingField.setRobot(robot, dir, x, y);
        assertRobotPosition(robot, dir, x, y);
    }

    private void assertRobotPosition(Robot robot, Direction dir, int x, int y) {
        assertEquals(String.format("Robot %d should be at location %s.", robot
                .getID(), new Point(x, y)), new Point(x, y), robot.getLocation());
        assertEquals(String.format("Robot %d should be facing %s.", robot
                .getID(), dir.toString()), dir, robot.getDirection());
    }

    private void testStandardPlayingField(int level) {
        PlayingField field = new PlayingField(4, level);
        assertEquals("The wrong level was returned", level, field.getLevel());
        assertCorrectDimensions(field, 51, 51);
        testStandardPlayingFieldGrid(level);
        assertAllCellsUnchecked(field);
        assertFalse("There should have been unclaimed buildings", field
                .allBuildingsClaimed());
    }

    private void assertCorrectDimensions(PlayingField field, int length,
            int width) {
        final String dimensionMessage = "The playing field's dimensions were wrong";
        assertEquals(dimensionMessage, length, field.getHeight());
        assertEquals(dimensionMessage, width, field.getWidth());
    }

    private void assertInitialRobotPositions(PlayingField field) {
        final Direction[] directions = { Direction.EAST, Direction.WEST,
                Direction.SOUTH, Direction.NORTH };
        int width = field.getWidth();
        int length = field.getHeight();
        final int[] xPos = { 0, width - 1, width - 1, 0 };
        final int[] yPos = { 0, length - 1, 0, length - 1 };
        List<Robot> robots = field.getRobots();
        for (int i = 0; i < robots.size() && i < 4; i++) {
            assertRobotPosition(robots.get(i), directions[i], xPos[i], yPos[i]);
        }
    }

    private void testStandardPlayingFieldGrid(int level) {
        PlayingField field = new PlayingField(4, level);
        testCommonPlayingFieldGrid(field);
        testLevel1Grid(field);
    }

    private void testCommonPlayingFieldGrid(PlayingField field) {
        for (int length = 0; length < field.getHeight(); length++) {
            for (int width = 0; width < field.getWidth(); width++) {
                Cell cell = field.getCell(width, length);
                if ((length % 2) == 1) // the length is odd, so we should find
                // alternate buildings and streets
                {
                    if ((width % 2) == 1) // the width is odd, so we should
                    // find a building
                    {
                        assertTrue("The cell at (" + width + ", " + length
                                + ") should be a building.",
                                cell instanceof BuildingCell);
                    }
                }
            }
        }
    }

    private void testLevel1Grid(PlayingField field) {
        for (int y = 0; y < field.getHeight(); y++) {
            for (int x = 0; x < field.getWidth(); x++) {
                Cell cell = field.getCell(x, y);
                if ((y % 2) == 1) { // the length is odd, so we should
                    // find
                    // alternate buildings and streets
                    if ((x % 2) == 0) { // the width is even, so we should
                        // find a street
                        assertTrue("The cell at (" + x + ", " + y
                                + ") should be a street.",
                                cell instanceof StreetCell);
                    } else {
                        String message = String
                                .format(
                                        "The cell at (%d, %d) should be a building cell",
                                        x, y);
                        assertTrue(message, cell instanceof BuildingCell);
                        assertEquals(
                                "The building cell should be worht 1 point", 1,
                                ((BuildingCell) cell).getPoints());
                    }
                } else { // the length is even, so we
                    // should find all streets this row
                    assertTrue("The cell at (" + x + ", " + y
                            + ") should be a street.",
                            cell instanceof StreetCell);
                }
            }
        }
    }

    private List<BuildingCell> testHigherLevelGrid(PlayingField field,
            int expectedRandomBuildings) {
        List<BuildingCell> cells = new ArrayList<BuildingCell>();

        int totalRandomBuildings = 0;
        for (int y = 0; y < field.getHeight(); y++) {
            for (int x = 0; x < field.getWidth(); x++) {
                Cell cell = field.getCell(x, y);
                if (cell instanceof BuildingCell) {
                    cells.add((BuildingCell) cell);
                }
                if (!(y % 2 == 1 && x % 2 == 1)) {
                    if (cell instanceof BuildingCell) {
                        totalRandomBuildings++;
                    }
                }
            }
        }
        assertEquals("There wasn't the correct number of random buildings",
                expectedRandomBuildings, totalRandomBuildings);
        return cells;
    }

    private void testGridForRandomness(int level, int expectedRandomBuildings) {
        final int numTests = 200;

        List<Integer> cells = new ArrayList<Integer>();
        List<Integer> fullSet = new ArrayList<Integer>();

        for (int i = 1; i <= 5; i++) {
            fullSet.add(i);
        }

        for (int i = 0; i < numTests; i++) {
            PlayingField field = new PlayingField(1, level);
            for (BuildingCell cell : testHigherLevelGrid(field,
                    expectedRandomBuildings)) {
                cells.add(cell.getPoints());
            }
        }

        TestUtils.verifyRandomness(cells, fullSet);
    }

    //TODO Write test to verify proper building positions.
}
