package backend;

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import junit.framework.TestCase;

public class RefereeTest extends TestCase {

	PlayingField _field;

    Referee _referee0;

    private Robot _robot0;
    private Robot _robot1;

    public void setUp() {
        List<Robot> robots = new ArrayList<Robot>();
        _robot0 = new Robot(0);
        _robot1 = new Robot(1);
        robots.add(_robot0);
        robots.add(_robot1);
        
        _field = new PlayingField(robots, 1, null);
        _referee0 = new Referee(_field, _robot0);

        _field.setRobot(_robot0, Direction.WEST, 0, 0);
    }
    
	public void testConstructor() {
		PlayingField field = new PlayingField(1, 1);
		Referee referee = new Referee(field, _robot1);
		assertFalse("The crumbs should be cleared initially", referee
				.crumbsCleared());
	}

	public void testAheadWithoutCrumbs() {
        createSimpleEnclosure(0, 0, 10, 10, _robot0, _referee0);
        setTrailOfCell(0, 8, _robot1, true);
        int pointsAwarded = _referee0.checkAndMakeMove(Move.AHEAD_WITHOUT_CRUMBS);
        assertEquals("The wrong points were awarded", 0, pointsAwarded);
        assertTrue("The crumbs should be cleared", _referee0.crumbsCleared());
        assertFalse("The crumbs cleared variable should be reset", _referee0
                .crumbsCleared());
        assertAllCrumbsCleared(_field, _robot0);
        assertTrue("The other robot's crumbs shouldn't be cleared",
                ((StreetCell) _field.getCell(0, 8)).hasRobotTrail(_robot1));
    }

	public void testAheadWithCrumbs() {
        Robot robot = _robot0;

        createSimpleEnclosure(0, 0, 10, 10, robot, _referee0);
        _robot0.setLocation(new Point(1, 0));
        setTrailOfCell(0, 8, _robot1, true);
        int pointsAwarded = _referee0.checkAndMakeMove(Move.AHEAD_WITH_CRUMBS);
        assertEquals("The wrong points were awarded", 25, pointsAwarded);
        assertTrue("The crumbs should be cleared", _referee0.crumbsCleared());
        assertFalse("The crumbs cleared variable should be reset", _referee0
                .crumbsCleared());
        assertAllCrumbsCleared(_field, robot);
        assertTrue("The other robot's crumbs shouldn't be cleared",
                ((StreetCell) _field.getCell(0, 8)).hasRobotTrail(_robot1));
    }

	public void testOtherMovementMoves() {
		createSimpleEnclosure(0, 0, 10, 10, _robot0, _referee0);
		setTrailOfCell(0, 8, _robot1, true);

		List<Point> allCellsInTrail = getCellsInTrail(_field, _robot0);

		Move[] moves = { Move.SKIP, Move.TURN_LEFT, Move.TURN_RIGHT };
		for (Move move : moves) {
            int pointsAwarded = _referee0.checkAndMakeMove(move);
			assertEquals("The wrong points were awarded", 0, pointsAwarded);
			assertTrue("The other robot's crumbs shouldn't be cleared",
					((StreetCell) _field.getCell(0, 8)).hasRobotTrail(_robot1));
			assertEquals("The robot's crumbs shouldn't be cleared",
					allCellsInTrail, getCellsInTrail(_field, _robot0));
		}
	}

	public void testMovingBackward() {
        createSimpleEnclosure(0, 0, 10, 10, _robot0, _referee0);
        setTrailOfCell(0, 8, _robot1, true);
        _field.setRobot(_robot0, Direction.EAST, 1, 0);
        int pointsAwarded = _referee0.checkAndMakeMove(Move.AHEAD_WITH_CRUMBS);
        assertEquals("The wrong points were awarded", 0, pointsAwarded);
        assertTrue("The crumbs should be cleared", _referee0.crumbsCleared());
        assertFalse("The crumbs cleared variable should be reset", _referee0
                .crumbsCleared());
        assertAllCrumbsClearedOtherThan(_field, _robot0, 1, 0);
        assertTrue("The other robot's crumbs shouldn't be cleared",
                ((StreetCell) _field.getCell(0, 8)).hasRobotTrail(_robot1));
    }

	private List<Point> getCellsInTrail(PlayingField field, Robot robot) {
		List<Point> result = new ArrayList<Point>();
		for (int x = 0; x < field.getWidth(); x++) {
			for (int y = 0; y < field.getHeight(); y++) {
				Cell cell = field.getCell(x, y);
				if (cell instanceof StreetCell) {
					StreetCell streetCell = (StreetCell) cell;
					if (streetCell.getRobotOccupying() == robot) {
						result.add(new Point(x, y));
					}
				}
			}
		}
		return result;
	}

	private void assertAllCrumbsCleared(PlayingField field, Robot robot) {
		for (int x = 0; x < field.getWidth(); x++) {
			for (int y = 0; y < field.getHeight(); y++) {
				Cell cell = field.getCell(x, y);
				if (cell instanceof StreetCell) {
					StreetCell streetCell = (StreetCell) cell;
					assertFalse(String.format(
							"The robot's crumbs should be cleared at (%d, %d)",
							x, y), streetCell.hasRobotTrail(robot));
				}
			}
		}
	}
	
	private void assertAllCrumbsClearedOtherThan(PlayingField field, Robot robot, int xPos, int yPos) {
		for (int x = 0; x < field.getWidth(); x++) {
			for (int y = 0; y < field.getHeight(); y++) {
				if (x == xPos && y == yPos) continue;
				Cell cell = field.getCell(x, y);
				if (cell instanceof StreetCell) {
					StreetCell streetCell = (StreetCell) cell;
					assertFalse(String.format(
							"The robot's crumbs should be cleared at (%d, %d)",
							x, y), streetCell.hasRobotTrail(robot));
				}
			}
		}
	}

	private void setTrailOfCell(int x, int y, Robot robot, boolean on) {
		((StreetCell) _field.getCell(x, y)).setRobotTrail(robot, on);
	}

	private boolean isCellCrumby(int x, int y, Robot robot) {
		return ((StreetCell) _field.getCell(x, y)).hasRobotTrail(robot);
	}

    private void createSimpleEnclosure(int xStart, int yStart, int width,
            int height, Robot robot, Referee referee) {
        _field.setRobot(robot, Direction.SOUTH, xStart, yStart);

        for (int y = yStart; y < yStart + height; y++) {
            _field.setRobot(robot, Direction.SOUTH, xStart, y);
            int score = _referee0.checkAndMakeMove(Move.AHEAD_WITH_CRUMBS);

            assertEquals(0, score);
            assertTrue("The crumbs weren't set properly", isCellCrumby(xStart,
                    y, robot));
        }

        for (int x = xStart; x < xStart + width; x++) {
            _field.setRobot(robot, Direction.EAST, x, yStart + height);
            int score = _referee0.checkAndMakeMove(Move.AHEAD_WITH_CRUMBS);
            assertEquals(0, score);
            assertTrue("The crumbs weren't set properly", isCellCrumby(x,
                    yStart + height, robot));
        }

        for (int y = yStart + height; y > 0; y--) {
            _field.setRobot(robot, Direction.NORTH, xStart + width, y);
            int score = _referee0.checkAndMakeMove(Move.AHEAD_WITH_CRUMBS);
            assertEquals(0, score);
            assertTrue("The crumbs weren't set properly", isCellCrumby(
                    xStart + width, y, robot));
        }

        for (int x = xStart + width; x > 1; x--) {
            _field.setRobot(robot, Direction.WEST, x, yStart);
            int score = _referee0.checkAndMakeMove(Move.AHEAD_WITH_CRUMBS);
            assertEquals(0, score);
            assertTrue("The crumbs weren't set properly", isCellCrumby(x,
                    yStart, robot));
        }
    }
    
    //TODO test robots other than #0.
}
